# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains base classes for the ledger API connection."""
import asyncio
from abc import ABC, abstractmethod
from asyncio import Task
from concurrent.futures._base import Executor
from logging import Logger
from typing import Any, Callable, Dict, Optional

from aea.configurations.base import PublicId
from aea.crypto.base import LedgerApi
from aea.crypto.registries import Registry, ledger_apis_registry
from aea.helpers.async_utils import AsyncState
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue, Dialogues


CONNECTION_ID = PublicId.from_str("fetchai/ledger:0.7.0")


class RequestDispatcher(ABC):
    """Base class for a request dispatcher."""

    TIMEOUT = 3
    MAX_ATTEMPTS = 120

    def __init__(
        self,
        logger: Logger,
        connection_state: AsyncState,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        executor: Optional[Executor] = None,
        api_configs: Optional[Dict[str, Dict[str, str]]] = None,
    ):
        """
        Initialize the request dispatcher.

        :param loop: the asyncio loop.
        :param executor: an executor.
        """
        self.connection_state = connection_state
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.executor = executor
        self._api_configs = api_configs
        self.logger = logger

    def api_config(self, ledger_id: str) -> Dict[str, str]:
        """Get api config."""
        config = {}  # type: Dict[str, str]
        if self._api_configs is not None and ledger_id in self._api_configs:
            config = self._api_configs[ledger_id]
        return config

    async def run_async(
        self,
        func: Callable[[Any], Task],
        api: LedgerApi,
        message: Message,
        dialogue: Dialogue,
    ):
        """
        Run a function in executor.

        :param func: the function to execute.
        :param args: the arguments to pass to the function.
        :return: the return value of the function.
        """
        try:
            response = await self.loop.run_in_executor(
                self.executor, func, api, message, dialogue
            )
            return response
        except Exception as e:  # pylint: disable=broad-except
            return self.get_error_message(e, api, message, dialogue)

    def dispatch(self, envelope: Envelope) -> Task:
        """
        Dispatch the request to the right sender handler.

        :param envelope: the envelope.
        :return: an awaitable.
        """
        if not isinstance(envelope.message, Message):  # pragma: nocover
            raise ValueError("Ledger connection expects non-serialized messages.")
        message = envelope.message
        ledger_id = self.get_ledger_id(message)
        api = self.ledger_api_registry.make(ledger_id, **self.api_config(ledger_id))
        dialogue = self.dialogues.update(message)
        if dialogue is None:
            raise ValueError(  # pragma: nocover
                "No dialogue created. Message={} not valid.".format(message)
            )
        performative = message.performative
        handler = self.get_handler(performative)
        return self.loop.create_task(self.run_async(handler, api, message, dialogue))

    def get_handler(self, performative: Any) -> Callable[[Any], Task]:
        """
        Get the handler method, given the message performative.

        :param performative: the message performative.
        :return: the method that will send the request.
        """
        handler = getattr(self, performative.value, None)
        if handler is None:
            raise Exception("Performative not recognized.")
        return handler

    @abstractmethod
    def get_error_message(
        self, e: Exception, api: LedgerApi, message: Message, dialogue: Dialogue,
    ) -> Message:
        """
        Build an error message.

        :param e: the exception
        :param api: the ledger api
        :param message: the received message.
        :param dialogue: the dialogue.
        :return: an error message response.
        """

    @property
    @abstractmethod
    def dialogues(self) -> Dialogues:
        """Get the dialogues."""

    @property
    def ledger_api_registry(self) -> Registry:
        """Get the registry."""
        return ledger_apis_registry

    @abstractmethod
    def get_ledger_id(self, message: Message) -> str:
        """Extract the ledger id from the message."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the ledger api registry."""

import logging

import pytest

import aea.crypto

from tests.conftest import (
    ETHEREUM,
    ETHEREUM_ADDRESS_ONE,
    ETHEREUM_TESTNET_CONFIG,
    FETCHAI,
    FETCHAI_ADDRESS_ONE,
    FETCHAI_TESTNET_CONFIG,
)


logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "identifier,address,config",
    [
        (FETCHAI, FETCHAI_ADDRESS_ONE, FETCHAI_TESTNET_CONFIG),
        (ETHEREUM, ETHEREUM_ADDRESS_ONE, ETHEREUM_TESTNET_CONFIG),
    ],
)
def test_make_ledger_apis(identifier, address, config):
    """Test the 'make' method for ledger api."""
    api = aea.crypto.registries.make_ledger_api(identifier, **config)

    # minimal functional test - comprehensive tests on ledger APIs are located in another module
    balance_1 = api.get_balance(address)
    balance_2 = api.get_balance(address)
    assert balance_1 == balance_2



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the agent module."""
import asyncio
from threading import Thread
from unittest.mock import Mock

import pytest

from aea.agent import Agent, Identity
from aea.runtime import RuntimeStates

from packages.fetchai.connections.local.connection import LocalNode

from tests.common.utils import wait_for_condition
from tests.conftest import _make_local_connection


class DummyAgent(Agent):
    """A dummy agent for testing."""

    def __init__(self, *args, **kwargs):
        """Initialize the agent."""
        super().__init__(*args, **kwargs)

    def setup(self) -> None:
        """Set up the agent."""
        pass

    def act(self) -> None:
        """Act."""
        pass

    def react(self) -> None:
        """React to events."""
        pass

    def update(self) -> None:
        """Update the state of the agent."""
        pass

    def teardown(self) -> None:
        """Tear down the agent."""
        pass

    @property
    def resources(self):
        """Get resources."""
        return Mock()


def test_run_agent():
    """Test that we can set up and then run the agent."""
    with LocalNode() as node:
        agent_name = "dummyagent"
        agent_address = "some_address"
        identity = Identity(agent_name, address=agent_address)
        oef_local_connection = _make_local_connection(agent_address, node)
        oef_local_connection._local_node = node

        agent = DummyAgent(
            identity, [oef_local_connection], loop=asyncio.new_event_loop()
        )
        agent_thread = Thread(target=agent.start)
        assert agent.state == RuntimeStates.stopped
        agent_thread.start()
        try:
            wait_for_condition(
                lambda: agent.state == RuntimeStates.running,
                timeout=10,
                error_msg="Agent state must be 'running'",
            )
        finally:
            agent.stop()
            assert agent.state == RuntimeStates.stopped
            agent_thread.join()


def test_runtime_modes():
    """Test runtime modes are set."""
    agent_name = "dummyagent"
    agent_address = "some_address"
    identity = Identity(agent_name, address=agent_address)
    agent = DummyAgent(identity, [],)

    assert not agent.is_running
    assert agent.is_stopped

    agent._runtime_mode = "not exists"

    with pytest.raises(ValueError):
        agent._get_runtime_class()

    with pytest.raises(ValueError):
        agent.runtime._get_main_loop_class("not exists")

    assert agent.runtime.loop_mode == agent.runtime.DEFAULT_RUN_LOOP



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helpers.search.generic."""

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.generic import GenericDataModel


def test_generic_data_model():
    """Test generic data model creation."""
    # ok
    GenericDataModel(
        "test", {"attr1": {"name": "attr1", "type": "str", "is_required": True}}
    )

    # bad type
    with pytest.raises(AEAEnforceError):
        GenericDataModel(
            "test",
            {"attr1": {"name": "attr1", "type": "bad type", "is_required": True}},
        )

    # bad name
    with pytest.raises(AEAEnforceError):
        GenericDataModel(
            "test", {"attr1": {"name": 1231, "type": "str", "is_required": True}}
        )

    # bad is required
    with pytest.raises(AEAEnforceError):
        GenericDataModel(
            "test", {"attr1": {"name": "attr1", "type": "str", "is_required": "True"}}
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea get_multiaddress' subcommand."""
import re
import typing
from pathlib import Path
from typing import Optional, Tuple, cast

import click
from click import ClickException

from aea.cli.utils.click_utils import PublicIdParameter
from aea.cli.utils.config import load_item_config
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project
from aea.cli.utils.package_utils import get_package_path_unified
from aea.configurations.base import ConnectionConfig, PublicId
from aea.crypto.base import Crypto
from aea.crypto.registries import crypto_registry
from aea.exceptions import enforce
from aea.helpers.multiaddr.base import MultiAddr


URI_REGEX = re.compile(r"(?:https?://)?(?P<host>[^:/ ]+):(?P<port>[0-9]*)")


@click.command()
@click.argument(
    "ledger_id",
    metavar="TYPE",
    type=click.Choice(list(crypto_registry.supported_ids)),
    required=True,
)
@click.option("-c", "--connection", is_flag=True)
@click.option(
    "-i", "--connection-id", type=PublicIdParameter(), required=False, default=None,
)
@click.option(
    "-h", "--host-field", type=str, required=False, default=None,
)
@click.option(
    "-p", "--port-field", type=str, required=False, default=None,
)
@click.option(
    "-u", "--uri-field", type=str, required=False, default="public_uri",
)
@click.pass_context
@check_aea_project
def get_multiaddress(
    click_context,
    ledger_id: str,
    connection: bool,
    connection_id: Optional[PublicId],
    host_field: str,
    port_field: str,
    uri_field: str,
):
    """Get the multiaddress associated with a private key or connection."""
    address = _try_get_multiaddress(
        click_context,
        ledger_id,
        connection,
        connection_id,
        host_field,
        port_field,
        uri_field,
    )
    click.echo(address)


def _try_get_multiaddress(
    click_context,
    ledger_id: str,
    is_connection: bool,
    connection_id: Optional[PublicId],
    host_field: str,
    port_field: str,
    uri_field: str,
):
    """
    Try to get the multi-address.

    :param click_context: click context object.
    :param ledger_id: the ledger id.
    :param is_connection: whether the key to load is from the wallet or from connections.
    :param connection_id: the connection id.
    :param host_field: if connection_id specified, the config field to retrieve the host
    :param port_field: if connection_id specified, the config field to retrieve the port

    :return: address.
    """
    ctx = cast(Context, click_context.obj)
    # connection_id not None implies is_connection
    is_connection = connection_id is not None or is_connection

    private_key_paths = (
        ctx.agent_config.private_key_paths
        if not is_connection
        else ctx.agent_config.connection_private_key_paths
    )
    private_key_path = private_key_paths.read(ledger_id)

    if private_key_path is None:
        raise ClickException(
            f"Cannot find '{ledger_id}'. Please check {'private_key_path' if not is_connection else 'connection_private_key_paths'}."
        )

    path_to_key = Path(private_key_path)
    crypto = crypto_registry.make(ledger_id, private_key_path=path_to_key)

    if connection_id is None:
        return _try_get_peerid(crypto)
    return _try_get_connection_multiaddress(
        click_context,
        crypto,
        cast(PublicId, connection_id),
        host_field,
        port_field,
        uri_field,
    )


def _try_get_peerid(crypto: Crypto) -> str:
    """Try to get the peer id."""
    try:
        peer_id = MultiAddr("", 0, crypto.public_key).peer_id
        return peer_id
    except Exception as e:
        raise ClickException(str(e))


def _read_host_and_port_from_config(
    connection_config: ConnectionConfig,
    uri_field: str,
    host_field: Optional[str],
    port_field: Optional[str],
) -> Tuple[str, int]:
    """
    Read host and port from config connection.

    :param host_field: the host field.
    :param port_field: the port field.
    :param uri_field: the uri field.
    :return: the host and the port.
    """
    host_is_none = host_field is None
    port_is_none = port_field is None
    one_is_none = (not host_is_none and port_is_none) or (
        host_is_none and not port_is_none
    )
    if not host_is_none and not port_is_none:
        if host_field not in connection_config.config:
            raise ClickException(
                f"Host field '{host_field}' not present in connection configuration {connection_config.public_id}"
            )
        if port_field not in connection_config.config:
            raise ClickException(
                f"Port field '{port_field}' not present in connection configuration {connection_config.public_id}"
            )
        host = connection_config.config[host_field]
        port = int(connection_config.config[port_field])
        return host, port
    if one_is_none:
        raise ClickException(
            "-h/--host-field and -p/--port-field must be specified together."
        )
    if uri_field not in connection_config.config:
        raise ClickException(
            f"URI field '{uri_field}' not present in connection configuration {connection_config.public_id}"
        )
    url_value = connection_config.config[uri_field]
    try:
        m = URI_REGEX.search(url_value)
        enforce(m is not None, f"URI Doesn't match regex '{URI_REGEX}'")
        m = cast(typing.Match, m)
        host = m.group("host")
        port = int(m.group("port"))
        return host, port
    except Exception as e:
        raise ClickException(
            f"Cannot extract host and port from {uri_field}: '{url_value}'. Reason: {str(e)}"
        )


def _try_get_connection_multiaddress(
    click_context,
    crypto: Crypto,
    connection_id: PublicId,
    host_field: Optional[str],
    port_field: Optional[str],
    uri_field: str,
) -> str:
    """
    Try to get the connection multiaddress.

    The host and the port options have the precedence over the uri option.

    :param click_context: the click context object.
    :param crypto: the crypto.
    :param connection_id: the connection id.
    :param host_field: the host field.
    :param port_field: the port field.
    :param uri_field: the uri field.
    :return: the multiaddress.
    """
    ctx = cast(Context, click_context.obj)
    if connection_id not in ctx.agent_config.connections:
        raise ValueError(f"Cannot find connection with the public id {connection_id}.")

    package_path = Path(get_package_path_unified(ctx, "connection", connection_id))
    connection_config = cast(
        ConnectionConfig, load_item_config("connection", package_path)
    )

    host, port = _read_host_and_port_from_config(
        connection_config, uri_field, host_field, port_field
    )

    try:
        multiaddr = MultiAddr(host, port, crypto.public_key)
        return multiaddr.format()
    except Exception as e:
        raise ClickException(f"An error occurred while creating the multiaddress: {e}")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea generate' subcommand."""

import os
from typing import cast

import click

from aea.cli.fingerprint import _fingerprint_item
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project, clean_after
from aea.cli.utils.loggers import logger
from aea.configurations.base import (
    DEFAULT_AEA_CONFIG_FILE,
    ProtocolSpecificationParseError,
    PublicId,
)
from aea.protocols.generator.base import ProtocolGenerator
from aea.protocols.generator.common import load_protocol_specification


@click.group()
@click.pass_context
@check_aea_project
def generate(click_context):  # pylint: disable=unused-argument
    """Generate a resource for the agent."""


@generate.command()
@click.argument("protocol_specification_path", type=str, required=True)
@click.pass_context
def protocol(click_context, protocol_specification_path: str):
    """Generate a protocol based on a specification and add it to the configuration file and agent."""
    _generate_item(click_context, "protocol", protocol_specification_path)


@clean_after
def _generate_item(click_context, item_type, specification_path):
    """Generate an item based on a specification and add it to the configuration file and agent."""
    ctx = cast(Context, click_context.obj)

    # Get existing items
    existing_id_list = getattr(ctx.agent_config, "{}s".format(item_type))
    existing_item_list = [public_id.name for public_id in existing_id_list]

    item_type_plural = item_type + "s"

    # Load item specification yaml file
    try:
        protocol_spec = load_protocol_specification(specification_path)
    except Exception as e:
        raise click.ClickException(str(e))

    protocol_directory_path = os.path.join(
        ctx.cwd, item_type_plural, protocol_spec.name
    )

    # Check if we already have an item with the same name in the agent config
    logger.debug(
        "{} already supported by the agent: {}".format(
            item_type_plural, existing_item_list
        )
    )
    if protocol_spec.name in existing_item_list:
        raise click.ClickException(
            "A {} with name '{}' already exists. Aborting...".format(
                item_type, protocol_spec.name
            )
        )
    # Check if we already have a directory with the same name in the resource directory (e.g. protocols) of the agent's directory
    if os.path.exists(protocol_directory_path):
        raise click.ClickException(
            "A directory with name '{}' already exists. Aborting...".format(
                protocol_spec.name
            )
        )

    ctx.clean_paths.append(protocol_directory_path)
    try:
        agent_name = ctx.agent_config.agent_name
        click.echo(
            "Generating {} '{}' and adding it to the agent '{}'...".format(
                item_type, protocol_spec.name, agent_name
            )
        )

        output_path = os.path.join(ctx.cwd, item_type_plural)
        protocol_generator = ProtocolGenerator(specification_path, output_path)
        warning_message = protocol_generator.generate()
        if warning_message is not None:
            click.echo(warning_message)

        # Add the item to the configurations
        logger.debug(
            "Registering the {} into {}".format(item_type, DEFAULT_AEA_CONFIG_FILE)
        )
        existing_id_list.add(
            PublicId(protocol_spec.author, protocol_spec.name, protocol_spec.version)
        )
        ctx.agent_loader.dump(
            ctx.agent_config, open(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w")
        )
    except FileExistsError:  # pragma: no cover
        raise click.ClickException(
            "A {} with this name already exists. Please choose a different name and try again.".format(
                item_type
            )
        )
    except ProtocolSpecificationParseError as e:
        raise click.ClickException(
            "The following error happened while parsing the protocol specification: "
            + str(e)
        )
    except Exception as e:
        raise click.ClickException(
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + str(e)
        )

    _fingerprint_item(click_context, "protocol", protocol_spec.public_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Portable pipe implementation for Linux, MacOS, and Windows."""

import asyncio
import errno
import logging
import os
import socket
import struct
import tempfile
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
from shutil import rmtree
from typing import IO, Optional

from aea.exceptions import enforce


_default_logger = logging.getLogger(__name__)

PIPE_CONN_TIMEOUT = 10.0
PIPE_CONN_ATTEMPTS = 10

TCP_SOCKET_PIPE_CLIENT_CONN_ATTEMPTS = 5


class IPCChannelClient(ABC):
    """Multi-platform interprocess communication channel for the client side."""

    @abstractmethod
    async def connect(self, timeout=PIPE_CONN_TIMEOUT) -> bool:
        """
        Connect to communication channel

        :param timeout: timeout for other end to connect
        """

    @abstractmethod
    async def write(self, data: bytes) -> None:
        """
        Write `data` bytes to the other end of the channel

        Will first write the size than the actual data

        :param data: bytes to write
        """

    @abstractmethod
    async def read(self) -> Optional[bytes]:
        """
        Read bytes from the other end of the channel

        Will first read the size than the actual data

        :return: read bytes
        """

    @abstractmethod
    async def close(self) -> None:
        """
        Close the communication channel.

        :return: None
        """


class IPCChannel(IPCChannelClient):
    """Multi-platform interprocess communication channel."""

    @property
    @abstractmethod
    def in_path(self) -> str:
        """
        Rendezvous point for incoming communication.

        :return: path
        """

    @property
    @abstractmethod
    def out_path(self) -> str:
        """
        Rendezvous point for outgoing communication.

        :return: path
        """


class PosixNamedPipeProtocol:
    """Posix named pipes async wrapper communication protocol."""

    def __init__(
        self,
        in_path: str,
        out_path: str,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """
        Initialize a new posix named pipe.

        :param in_path: rendezvous point for incoming data
        :param out_path: rendezvous point for outgoing daa
        """

        self.logger = logger
        self._loop = loop
        self._in_path = in_path
        self._out_path = out_path
        self._in = -1
        self._out = -1

        self._stream_reader = None  # type: Optional[asyncio.StreamReader]
        self._reader_protocol = None  # type: Optional[asyncio.StreamReaderProtocol]
        self._fileobj = None  # type: Optional[IO[str]]

        self._connection_attempts = PIPE_CONN_ATTEMPTS
        self._connection_timeout = PIPE_CONN_TIMEOUT

    async def connect(self, timeout: float = PIPE_CONN_TIMEOUT) -> bool:
        """
        Connect to the other end of the pipe

        :param timeout: timeout before failing
        :return: connection success
        """

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._connection_timeout = timeout / PIPE_CONN_ATTEMPTS if timeout > 0 else 0
        if self._connection_attempts <= 1:  # pragma: no cover
            return False
        self._connection_attempts -= 1

        self.logger.debug(
            "Attempt opening pipes {}, {}...".format(self._in_path, self._out_path)
        )

        self._in = os.open(self._in_path, os.O_RDONLY | os.O_NONBLOCK | os.O_SYNC)

        try:
            self._out = os.open(self._out_path, os.O_WRONLY | os.O_NONBLOCK)
        except OSError as e:
            if e.errno == errno.ENXIO:
                self.logger.debug("Sleeping for {}...".format(self._connection_timeout))
                await asyncio.sleep(self._connection_timeout)
                return await self.connect(timeout)
            raise e  # pragma: no cover

        # setup reader
        enforce(
            self._in != -1 and self._out != -1 and self._loop is not None,
            "Incomplete initialization.",
        )
        self._stream_reader = asyncio.StreamReader(loop=self._loop)
        self._reader_protocol = asyncio.StreamReaderProtocol(
            self._stream_reader, loop=self._loop
        )
        self._fileobj = os.fdopen(self._in, "r")
        await self._loop.connect_read_pipe(
            lambda: self.__reader_protocol, self._fileobj
        )

        return True

    @property
    def __reader_protocol(self) -> asyncio.StreamReaderProtocol:
        """Get reader protocol."""
        if self._reader_protocol is None:
            raise ValueError("reader protocol not set!")  # pragma: nocover
        return self._reader_protocol

    async def write(self, data: bytes) -> None:
        """
        Write to pipe.

        :param data: bytes to write to pipe
        """
        self.logger.debug("writing {}...".format(len(data)))
        size = struct.pack("!I", len(data))
        os.write(self._out, size + data)
        await asyncio.sleep(0.0)

    async def read(self) -> Optional[bytes]:
        """
        Read from pipe.

        :return: read bytes
        """
        if self._stream_reader is None:  # pragma: nocover
            raise ValueError("StreamReader not set, call connect first!")
        try:
            self.logger.debug("waiting for messages (in={})...".format(self._in_path))
            buf = await self._stream_reader.readexactly(4)
            if not buf:  # pragma: no cover
                return None
            size = struct.unpack("!I", buf)[0]
            if size <= 0:  # pragma: no cover
                return None
            data = await self._stream_reader.readexactly(size)
            if not data:  # pragma: no cover
                return None
            return data
        except asyncio.IncompleteReadError as e:  # pragma: no cover
            self.logger.info(
                "Connection disconnected while reading from pipe ({}/{})".format(
                    len(e.partial), e.expected
                )
            )
            return None
        except asyncio.CancelledError:  # pragma: no cover
            return None

    async def close(self) -> None:
        """Disconnect pipe."""
        self.logger.debug("closing pipe (in={})...".format(self._in_path))
        if self._fileobj is None:
            raise ValueError("Pipe not connected")  # pragma: nocover
        try:
            # TOFIX(LR) Hack for MacOSX
            size = struct.pack("!I", 0)
            os.write(self._out, size)

            os.close(self._out)
            self._fileobj.close()
        except OSError:  # pragma: no cover
            pass
        await asyncio.sleep(0)


class TCPSocketProtocol:
    """TCP socket communication protocol."""

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """
        Initialize the tcp socket protocol.

        :param reader: established asyncio reader
        :param writer: established asyncio writer
        """

        self.logger = logger
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self._reader = reader
        self._writer = writer

    async def write(self, data: bytes) -> None:
        """
        Write to socket.

        :param data: bytes to write
        """
        if self._writer is None:
            raise ValueError("writer not set!")  # pragma: nocover
        self.logger.debug("writing {}...".format(len(data)))
        size = struct.pack("!I", len(data))
        self._writer.write(size + data)
        await self._writer.drain()

    async def read(self) -> Optional[bytes]:
        """
        Read from socket.

        :return: read bytes
        """
        try:
            self.logger.debug("waiting for messages...")
            buf = await self._reader.readexactly(4)
            if not buf:  # pragma: no cover
                return None
            size = struct.unpack("!I", buf)[0]
            data = await self._reader.readexactly(size)
            if not data:  # pragma: no cover
                return None
            return data
        except asyncio.IncompleteReadError as e:  # pragma: no cover
            self.logger.info(
                "Connection disconnected while reading from pipe ({}/{})".format(
                    len(e.partial), e.expected
                )
            )
            return None
        except asyncio.CancelledError:  # pragma: no cover
            return None

    async def close(self) -> None:
        """Disconnect socket."""
        self._writer.write_eof()
        await self._writer.drain()
        self._writer.close()


class TCPSocketChannel(IPCChannel):
    """Interprocess communication channel implementation using tcp sockets."""

    def __init__(
        self,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """Initialize tcp socket interprocess communication channel."""
        self.logger = logger
        self._loop = loop
        self._server = None  # type: Optional[asyncio.AbstractServer]
        self._connected = None  # type: Optional[asyncio.Event]
        self._sock = None  # type: Optional[TCPSocketProtocol]

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        s.listen(1)
        self._port = s.getsockname()[1]
        s.close()

    async def connect(self, timeout: float = PIPE_CONN_TIMEOUT) -> bool:
        """
        Setup communication channel and wait for other end to connect.

        :param timeout: timeout for the connection to be established
        """

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._connected = asyncio.Event()
        self._server = await asyncio.start_server(
            self._handle_connection, host="127.0.0.1", port=self._port
        )
        if self._server.sockets is None:
            raise ValueError("Server sockets is None!")  # pragma: nocover
        self._port = self._server.sockets[0].getsockname()[1]
        self.logger.debug("socket pipe rdv point: {}".format(self._port))

        try:
            await asyncio.wait_for(self._connected.wait(), timeout)
        except asyncio.TimeoutError:  # pragma: no cover
            return False

        self._server.close()
        await self._server.wait_closed()

        return True

    async def _handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        """Handle connection."""
        if self._connected is None:
            raise ValueError("Connected is None!")  # pragma: nocover
        self._connected.set()
        self._sock = TCPSocketProtocol(
            reader, writer, logger=self.logger, loop=self._loop
        )

    async def write(self, data: bytes) -> None:
        """
        Write to channel.

        :param data: bytes to write
        """
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        await self._sock.write(data)

    async def read(self) -> Optional[bytes]:
        """
        Read from channel.

        :param data: read bytes
        """
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        return await self._sock.read()

    async def close(self) -> None:
        """Disconnect from channel and clean it up."""
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        await self._sock.close()

    @property
    def in_path(self) -> str:
        """Rendezvous point for incoming communication."""
        return str(self._port)

    @property
    def out_path(self) -> str:
        """Rendezvous point for outgoing communication."""
        return str(self._port)


class PosixNamedPipeChannel(IPCChannel):
    """Interprocess communication channel implementation using Posix named pipes."""

    def __init__(
        self,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """Initialize posix named pipe interprocess communication channel."""
        self.logger = logger
        self._loop = loop

        self._pipe_dir = tempfile.mkdtemp()
        self._in_path = "{}/process_to_aea".format(self._pipe_dir)
        self._out_path = "{}/aea_to_process".format(self._pipe_dir)

        # setup fifos
        self.logger.debug(
            "Creating pipes ({}, {})...".format(self._in_path, self._out_path)
        )
        if os.path.exists(self._in_path):
            os.remove(self._in_path)  # pragma: no cover
        if os.path.exists(self._out_path):
            os.remove(self._out_path)  # pragma: no cover
        os.mkfifo(self._in_path)
        os.mkfifo(self._out_path)

        self._pipe = PosixNamedPipeProtocol(
            self._in_path, self._out_path, logger=logger, loop=loop
        )

    async def connect(self, timeout: float = PIPE_CONN_TIMEOUT) -> bool:
        """
        Setup communication channel and wait for other end to connect.

        :param timeout: timeout for connection to be established
        :return: bool, indicating sucess
        """

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        return await self._pipe.connect(timeout)

    async def write(self, data: bytes) -> None:
        """
        Write to the channel.

        :param data: data to write to channel
        """
        await self._pipe.write(data)

    async def read(self) -> Optional[bytes]:
        """
        Read from the channel.

        :return: read bytes
        """
        return await self._pipe.read()

    async def close(self) -> None:
        """Close the channel and clean it up."""
        await self._pipe.close()
        rmtree(self._pipe_dir)

    @property
    def in_path(self) -> str:
        """Rendezvous point for incoming communication."""
        return self._in_path

    @property
    def out_path(self) -> str:
        """Rendezvous point for outgoing communication."""
        return self._out_path


class TCPSocketChannelClient(IPCChannelClient):
    """Interprocess communication channel client using tcp sockets."""

    def __init__(  # pylint: disable=unused-argument
        self,
        in_path: str,
        out_path: str,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """
        Initialize a tcp socket communication channel client.

        :param in_path: rendezvous point for incoming data
        :param out_path: rendezvous point for outgoing data
        """
        self.logger = logger
        self._loop = loop

        self._port = int(in_path)
        self._sock = None  # type: Optional[TCPSocketProtocol]

        self._attempts = TCP_SOCKET_PIPE_CLIENT_CONN_ATTEMPTS
        self._timeout = PIPE_CONN_TIMEOUT / self._attempts

    async def connect(self, timeout: float = PIPE_CONN_TIMEOUT) -> bool:
        """
        Connect to the other end of the communication channel.

        :param timeout: timeout for connection to be established
        """
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._timeout = timeout / TCP_SOCKET_PIPE_CLIENT_CONN_ATTEMPTS

        self.logger.debug(
            "Attempting to connect to {}:{}.....".format("127.0.0.1", self._port)
        )

        connected = False
        while self._attempts > 0:
            self._attempts -= 1
            try:
                reader, writer = await asyncio.open_connection(
                    "127.0.0.1",
                    self._port,  # pylint: disable=protected-access
                    loop=self._loop,
                )
                self._sock = TCPSocketProtocol(
                    reader, writer, logger=self.logger, loop=self._loop
                )
                connected = True
                break
            except ConnectionRefusedError:
                await asyncio.sleep(self._timeout)
            except Exception:  # pylint: disable=broad-except  # pragma: nocover
                return False

        return connected

    async def write(self, data: bytes) -> None:
        """
        Write data to channel.

        :param data: bytes to write
        """
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        await self._sock.write(data)

    async def read(self) -> Optional[bytes]:
        """
        Read data from channel.

        :return: read bytes
        """
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        return await self._sock.read()

    async def close(self) -> None:
        """Disconnect from communication channel."""
        if self._sock is None:
            raise ValueError("Socket pipe not connected.")  # pragma: nocover
        await self._sock.close()


class PosixNamedPipeChannelClient(IPCChannelClient):
    """Interprocess communication channel client using Posix named pipes."""

    def __init__(
        self,
        in_path: str,
        out_path: str,
        logger: logging.Logger = _default_logger,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """
        Initialize a posix named pipe communication channel client.

        :param in_path: rendezvous point for incoming data
        :param out_path: rendezvous point for outgoing data
        """

        self.logger = logger
        self._loop = loop

        self._in_path = in_path
        self._out_path = out_path
        self._pipe = None  # type: Optional[PosixNamedPipeProtocol]

    async def connect(self, timeout: float = PIPE_CONN_TIMEOUT) -> bool:
        """
        Connect to the other end of the communication channel.

        :param timeout: timeout for connection to be established
        """

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._pipe = PosixNamedPipeProtocol(
            self._in_path, self._out_path, logger=self.logger, loop=self._loop
        )
        return await self._pipe.connect()

    async def write(self, data: bytes) -> None:
        """
        Write data to channel.

        :param data: bytes to write
        """
        if self._pipe is None:
            raise ValueError("Pipe not connected.")  # pragma: nocover
        await self._pipe.write(data)

    async def read(self) -> Optional[bytes]:
        """
        Read data from channel.

        :return: read bytes
        """
        if self._pipe is None:
            raise ValueError("Pipe not connected.")  # pragma: nocover
        return await self._pipe.read()

    async def close(self) -> None:
        """Disconnect from communication channel."""
        if self._pipe is None:
            raise ValueError("Pipe not connected.")  # pragma: nocover
        return await self._pipe.close()


def make_ipc_channel(
    logger: logging.Logger = _default_logger, loop: Optional[AbstractEventLoop] = None
) -> IPCChannel:
    """
    Build a portable bidirectional InterProcess Communication channel

    :param logger: the logger
    :param loop: the loop
    :return: IPCChannel
    """
    if os.name == "posix":
        return PosixNamedPipeChannel(logger=logger, loop=loop)
    if os.name == "nt":  # pragma: nocover
        return TCPSocketChannel(logger=logger, loop=loop)
    raise Exception(  # pragma: nocover
        "make ipc channel is not supported on platform {}".format(os.name)
    )


def make_ipc_channel_client(
    in_path: str,
    out_path: str,
    logger: logging.Logger = _default_logger,
    loop: Optional[AbstractEventLoop] = None,
) -> IPCChannelClient:
    """
    Build a portable bidirectional InterProcess Communication client channel

    :param in_path: rendezvous point for incoming communication
    :param out_path: rendezvous point for outgoing outgoing
    :param logger: the logger
    :param loop: the loop
    :return: IPCChannel
    """
    if os.name == "posix":
        return PosixNamedPipeChannelClient(in_path, out_path, logger=logger, loop=loop)
    if os.name == "nt":  # pragma: nocover
        return TCPSocketChannelClient(in_path, out_path, logger=logger, loop=loop)
    raise Exception(  # pragma: nocover
        "make ip channel client is not supported on platform {}".format(os.name)
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains a checker for PyPI version consistency."""
import operator
from collections import defaultdict
from typing import Dict, List, Set, cast

from packaging.specifiers import Specifier, SpecifierSet
from packaging.version import InvalidVersion, Version

from aea.configurations.base import Dependencies, Dependency


def and_(s1: SpecifierSet, s2: SpecifierSet):
    """Do the and between two specifier sets."""
    return operator.and_(s1, s2)


def _handle_compatibility_operator(
    all_specifiers: List[Specifier],
    operator_to_specifiers: Dict[str, Set[Specifier]],
    specifier: Specifier,
) -> None:
    """
    Handle a specifier with operator '~='.

    Split specifier of the form "~=<major.minor>" in two specifiers:
    - >= <major.minor>
    - < <major+1>
    Also handle micro-numbers, i.e. "~=<major.minor.micro>"
    (see last examples of https://www.python.org/dev/peps/pep-0440/#compatible-release)

    :param all_specifiers: the list of all specifiers (to be populated).
    :param operator_to_specifiers: a mapping from operator to specifiers (to be populated).
    :param specifier: the specifier to process.
    :return: None
    """
    spec_version = Version(specifier.version)
    base_version = spec_version.base_version
    parts = base_version.split(".")
    index_to_update = -2
    if (
        spec_version.is_prerelease
        or spec_version.is_devrelease
        or spec_version.is_postrelease
    ):
        # if it is a pre-release, ignore the suffix.
        index_to_update += 1
        parts = parts[:-1]
    # bump second-to-last part
    parts[index_to_update] = str(int(parts[index_to_update]) + 1)
    upper_version = Version(".".join(parts))
    spec_1 = Specifier(">=" + str(spec_version))
    spec_2 = Specifier("<" + str(upper_version))
    all_specifiers.extend([spec_1, spec_2])
    operator_to_specifiers[spec_1.operator].add(spec_1)
    operator_to_specifiers[spec_2.operator].add(spec_2)


def is_satisfiable(specifier_set: SpecifierSet) -> bool:
    """
    Check if the specifier set is satisfiable.

    Satisfiable means that there exists a version number
    that satisfies all the constraints. It is worth
    noticing that it doesn't mean that that version
    number with that package actually exists.

    >>> from packaging.specifiers import SpecifierSet

    The specifier set ">0.9, ==1.0" is satisfiable:
    the version number "1.0" satisfies the constraints

    >>> s1 = SpecifierSet(">0.9,==1.0")
    >>> "1.0" in s1
    True
    >>> is_satisfiable(s1)
    True

    The specifier set "==1.0, >1.1" is not satisfiable:

    >>> s1 = SpecifierSet("==1.0,>1.1")
    >>> is_satisfiable(s1)
    False

    For other details, please refer to PEP440:

        https://www.python.org/dev/peps/pep-0440

    :param specifier_set: the specifier set.
    :return: False if the constraints are surely non-satisfiable, True if we don't know.
    """
    # group single specifiers by operator
    all_specifiers: List[Specifier] = []
    operator_to_specifiers: Dict[str, Set[Specifier]] = defaultdict(set)
    # pre-processing
    for specifier in list(specifier_set):
        specifier = cast(Specifier, specifier)
        try:
            # if we can't parse the version number, ignore that specifier.
            # Even if it follows a legacy version format (e.g. "2.*")
            Version(specifier.version)
        except InvalidVersion:
            continue

        # handle specifiers with '~=' operators.
        if specifier.operator == "~=":
            _handle_compatibility_operator(
                all_specifiers, operator_to_specifiers, specifier
            )
        else:
            all_specifiers.append(specifier)
            operator_to_specifiers[specifier.operator].add(specifier)

    # end of pre-processing. Start evaluation
    # if there are two different "==" specifier, return False
    if len(operator_to_specifiers["=="]) >= 2:
        return False

    if len(operator_to_specifiers["=="]) == 1:
        eq_specifier = operator_to_specifiers["=="].pop()
        eq_version = eq_specifier.version
        # notice: we implicitly handle the "!=" constraints.
        return eq_version in specifier_set

    # group all the "<" or "<=" together. They are interpreted as conjunction.
    # simplify: the spec with the lowest version number is the strictest constraint
    less_than_strict_specs = operator_to_specifiers["<"]
    less_than_equal_specs = operator_to_specifiers["<="]
    less_than_specs = set.union(less_than_equal_specs, less_than_strict_specs)
    # sort less-than constraints in the following way: ["<1.0", "<=1.0", "<2.0", "<=2.0"]
    # the first element is the strictest constraint.
    sorted_less_than_specs = sorted(
        less_than_specs, key=lambda x: (Version(x.version), len(x.operator))
    )
    lowest_less_than = (
        sorted_less_than_specs[0] if len(sorted_less_than_specs) > 0 else None
    )

    # group all the ">" or ">=" together. They are interpreted as conjunction.
    # simplify: the spec with the greatest version number is the strictest constraint
    greater_than_strict_specs = operator_to_specifiers[">"]
    greater_than_equal_specs = operator_to_specifiers[">="]
    greater_than_specs = set.union(greater_than_strict_specs, greater_than_equal_specs)
    # sort greater-than constraints in the following way: [">=1.0", ">1.0", ">=2.0", ">2.0"]
    # the last element is the strictest constraint.
    sorted_greater_than_specs = sorted(
        greater_than_specs, key=lambda x: (Version(x.version), -len(x.operator))
    )
    greatest_greater_than = (
        sorted_greater_than_specs[-1] if len(sorted_greater_than_specs) > 0 else None
    )

    # if there exist two range constraints, check satisfiability.
    # otherwise, we can't say much.
    if lowest_less_than is not None and greatest_greater_than is not None:
        return _handle_range_constraints(lowest_less_than, greatest_greater_than)

    return True


def _handle_range_constraints(
    lowest_less_than: Specifier, greatest_greater_than: Specifier
) -> bool:
    """
    Check whether two specifiers of the following type are compatible.

    It is a helper method for the is_satisfiable function and checks:
    - "<=<some-version-number>"
    - ">=<some-version-number>"

    The equality might be optional.

    The pseudo-code is the following:
    - is the version number of the lower-than constraint smaller than the one of
      the greater-than constraint? E.g. we are in cases like "<=1.0" and ">1.1".
      -  If yes, then the constraints are unsatisfiable.
      -  Otherwise, they are satisfiable.
    - are the version numbers the same? E.g. "<=1.0,>=1.0"
      - If yes, if at least one of them is a strict comparison, then return False, otherwise True.
    - otherwise, return True.

    :param lowest_less_than: the less-than constraint.
    :param greatest_greater_than: the greater than constraint.
    :return: False if we are sure the two constraints are not satisfiable, True if we don't know.
    """
    version_less_than = Version(lowest_less_than.version)
    version_greater_than = Version(greatest_greater_than.version)
    if version_less_than < version_greater_than:
        return False
    if version_greater_than == version_less_than:
        # check if one of them has NOT the equality
        one_of_them_is_a_strict_comparison = (
            greatest_greater_than.operator == ">"
        ) or (lowest_less_than.operator == "<")
        return not one_of_them_is_a_strict_comparison
    return True


def is_simple_dep(dep: Dependency) -> bool:
    """
    Check if it is a simple dependency.

    Namely, if it has no field specified, or only the 'version' field set.

    :param dep: the dependency
    :return: whether it is a simple dependency or not
    """
    return dep.index is None and dep.git is None


def to_set_specifier(dep: Dependency) -> SpecifierSet:
    """Get the set specifier. It assumes to be a simple dependency (see above)."""
    return SpecifierSet(dep.version)


def merge_dependencies(dep1: Dependencies, dep2: Dependencies) -> Dependencies:
    """
    Merge two groups of dependencies.

    If some of them are not "simple" (see above), we just filter them out.

    :param dep1: the first operand
    :param dep2: the second operand.
    :return: the merged dependencies.
    """
    result: Dependencies
    result = {pkg_name: info for pkg_name, info in dep1.items() if is_simple_dep(info)}

    for pkg_name, info in dep2.items():
        if not is_simple_dep(info):
            continue
        new_specifier = SpecifierSet(info.version)
        old_specifier = (
            SpecifierSet(result[pkg_name].version)
            if pkg_name in result
            else SpecifierSet("")
        )
        combined_specifier = and_(new_specifier, old_specifier)
        new_info = Dependency(
            name=info.name,
            version=combined_specifier,
            index=info.index,
            git=info.git,
            ref=info.ref,
        )
        result[pkg_name] = new_info

    return result



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the protocol generator."""
# pylint: skip-file

import itertools
import os
import shutil
from datetime import date
from pathlib import Path
from typing import Optional, Tuple

from aea.protocols.generator.common import (
    CUSTOM_TYPES_DOT_PY_FILE_NAME,
    DIALOGUE_DOT_PY_FILE_NAME,
    INIT_FILE_NAME,
    MESSAGE_DOT_PY_FILE_NAME,
    MESSAGE_IMPORT,
    PATH_TO_PACKAGES,
    PROTOCOL_YAML_FILE_NAME,
    PYTHON_TYPE_TO_PROTO_TYPE,
    SERIALIZATION_DOT_PY_FILE_NAME,
    SERIALIZER_IMPORT,
    _camel_case_to_snake_case,
    _create_protocol_file,
    _get_sub_types_of_compositional_types,
    _includes_custom_type,
    _python_pt_or_ct_type_to_proto_type,
    _to_camel_case,
    _union_sub_type_to_protobuf_variable_name,
    check_prerequisites,
    check_protobuf_using_protoc,
    load_protocol_specification,
    try_run_black_formatting,
    try_run_isort_formatting,
    try_run_protoc,
)
from aea.protocols.generator.extract_specification import extract


def _copyright_header_str(author: str) -> str:
    """
    Produce the copyright header text for a protocol.

    :param author: the author of the protocol.
    :return: The copyright header text.
    """
    copy_right_str = (
        "# -*- coding: utf-8 -*-\n"
        "# ------------------------------------------------------------------------------\n"
        "#\n"
    )
    copy_right_str += "#   Copyright {} {}\n".format(date.today().year, author)
    copy_right_str += (
        "#\n"
        '#   Licensed under the Apache License, Version 2.0 (the "License");\n'
        "#   you may not use this file except in compliance with the License.\n"
        "#   You may obtain a copy of the License at\n"
        "#\n"
        "#       http://www.apache.org/licenses/LICENSE-2.0\n"
        "#\n"
        "#   Unless required by applicable law or agreed to in writing, software\n"
        '#   distributed under the License is distributed on an "AS IS" BASIS,\n'
        "#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n"
        "#   See the License for the specific language governing permissions and\n"
        "#   limitations under the License.\n"
        "#\n"
        "# ------------------------------------------------------------------------------\n"
    )
    return copy_right_str


class ProtocolGenerator:
    """This class generates a protocol_verification package from a ProtocolTemplate object."""

    def __init__(
        self,
        path_to_protocol_specification: str,
        output_path: str = ".",
        dotted_path_to_protocol_package: Optional[str] = None,
    ) -> None:
        """
        Instantiate a protocol generator.

        :param path_to_protocol_specification: path to protocol specification file
        :param output_path: the path to the location in which the protocol module is to be generated.
        :param dotted_path_to_protocol_package: the path to the protocol package

        :return: None
        """
        # Check the prerequisite applications are installed
        try:
            check_prerequisites()
        except FileNotFoundError:
            raise

        # Load protocol specification
        try:
            self.protocol_specification = load_protocol_specification(
                path_to_protocol_specification
            )
        except Exception:
            raise

        # Helper fields
        self.path_to_protocol_specification = path_to_protocol_specification
        self.protocol_specification_in_camel_case = _to_camel_case(
            self.protocol_specification.name
        )
        self.path_to_generated_protocol_package = os.path.join(
            output_path, self.protocol_specification.name
        )
        self.dotted_path_to_protocol_package = (
            dotted_path_to_protocol_package + self.protocol_specification.name
            if dotted_path_to_protocol_package is not None
            else "{}.{}.protocols.{}".format(
                PATH_TO_PACKAGES,
                self.protocol_specification.author,
                self.protocol_specification.name,
            )
        )
        self.indent = ""

        # Extract specification fields
        try:
            self.spec = extract(self.protocol_specification)
        except Exception:
            raise

    def _change_indent(self, number: int, mode: str = None) -> None:
        """
        Update the value of 'indent' global variable.

        This function controls the indentation of the code produced throughout the generator.

        There are two modes:
        - Setting the indent to a desired 'number' level. In this case, 'mode' has to be set to "s".
        - Updating the incrementing/decrementing the indentation level by 'number' amounts. In this case 'mode' is None.

        :param number: the number of indentation levels to set/increment/decrement
        :param mode: the mode of indentation change

        :return: None
        """
        if mode and mode == "s":
            if number >= 0:
                self.indent = number * "    "
            else:
                raise ValueError("Error: setting indent to be a negative number.")
        else:
            if number >= 0:
                for _ in itertools.repeat(None, number):
                    self.indent += "    "
            else:
                if abs(number) <= len(self.indent) / 4:
                    self.indent = self.indent[abs(number) * 4 :]
                else:
                    raise ValueError(
                        "Not enough spaces in the 'indent' variable to remove."
                    )

    def _import_from_typing_module(self) -> str:
        """
        Manage import statement for the typing package.

        :return: import statement for the typing package
        """
        ordered_packages = [
            "Dict",
            "FrozenSet",
            "Optional",
            "Set",
            "Tuple",
            "Union",
            "cast",
        ]
        import_str = "from typing import "
        for package in ordered_packages:
            if self.spec.typing_imports[package]:
                import_str += "{}, ".format(package)
        import_str = import_str[:-2]
        return import_str

    def _import_from_custom_types_module(self) -> str:
        """
        Manage import statement from custom_types module.

        :return: import statement for the custom_types module
        """
        import_str = ""
        if len(self.spec.all_custom_types) == 0:
            pass
        else:
            for custom_class in self.spec.all_custom_types:
                import_str += "from {}.custom_types import {} as Custom{}\n".format(
                    self.dotted_path_to_protocol_package, custom_class, custom_class,
                )
            import_str = import_str[:-1]
        return import_str

    def _performatives_str(self) -> str:
        """
        Generate the performatives instance property string, a set containing all valid performatives of this protocol.

        :return: the performatives set string
        """
        performatives_str = "{"
        for performative in self.spec.all_performatives:
            performatives_str += '"{}", '.format(performative)
        performatives_str = performatives_str[:-2]
        performatives_str += "}"
        return performatives_str

    def _performatives_enum_str(self) -> str:
        """
        Generate the performatives Enum class.

        :return: the performatives Enum string
        """
        enum_str = self.indent + "class Performative(Message.Performative):\n"
        self._change_indent(1)
        enum_str += self.indent + '"""Performatives for the {} protocol."""\n\n'.format(
            self.protocol_specification.name
        )
        for performative in self.spec.all_performatives:
            enum_str += self.indent + '{} = "{}"\n'.format(
                performative.upper(), performative
            )
        enum_str += "\n"
        enum_str += self.indent + "def __str__(self):\n"
        self._change_indent(1)
        enum_str += self.indent + '"""Get the string representation."""\n'
        enum_str += self.indent + "return str(self.value)\n"
        self._change_indent(-1)
        enum_str += "\n"
        self._change_indent(-1)

        return enum_str

    def _to_custom_custom(self, content_type: str) -> str:
        """
        Evaluate whether a content type is a custom type or has a custom type as a sub-type.

        :param content_type: the content type.
        :return: Boolean result
        """
        new_content_type = content_type
        if _includes_custom_type(content_type):
            for custom_type in self.spec.all_custom_types:
                new_content_type = new_content_type.replace(
                    custom_type, self.spec.custom_custom_types[custom_type]
                )
        return new_content_type

    def _check_content_type_str(self, content_name: str, content_type: str) -> str:
        """
        Produce the checks of elements of compositional types.

        :param content_name: the name of the content to be checked
        :param content_type: the type of the content to be checked

        :return: the string containing the checks.
        """
        check_str = ""
        if content_type.startswith("Optional["):
            optional = True
            check_str += self.indent + 'if self.is_set("{}"):\n'.format(content_name)
            self._change_indent(1)
            check_str += self.indent + "expected_nb_of_contents += 1\n"
            content_type = _get_sub_types_of_compositional_types(content_type)[0]
            check_str += self.indent + "{} = cast({}, self.{})\n".format(
                content_name, self._to_custom_custom(content_type), content_name
            )
            content_variable = content_name
        else:
            optional = False
            content_variable = "self." + content_name
        if content_type.startswith("Union["):
            element_types = _get_sub_types_of_compositional_types(content_type)
            unique_standard_types_set = set()
            for typing_content_type in element_types:
                if typing_content_type.startswith("FrozenSet"):
                    unique_standard_types_set.add("frozenset")
                elif typing_content_type.startswith("Tuple"):
                    unique_standard_types_set.add("tuple")
                elif typing_content_type.startswith("Dict"):
                    unique_standard_types_set.add("dict")
                else:
                    unique_standard_types_set.add(typing_content_type)
            unique_standard_types_list = sorted(unique_standard_types_set)
            check_str += self.indent
            check_str += "enforce("
            for unique_type in unique_standard_types_list:
                check_str += "type({}) == {} or ".format(
                    content_variable, self._to_custom_custom(unique_type)
                )
            check_str = check_str[:-4]
            check_str += ", \"Invalid type for content '{}'. Expected either of '{}'. Found '{{}}'.\".format(type({})))\n".format(
                content_name,
                [
                    unique_standard_type
                    for unique_standard_type in unique_standard_types_list
                ],
                content_variable,
            )
            if "frozenset" in unique_standard_types_list:
                check_str += self.indent + "if type({}) == frozenset:\n".format(
                    content_variable
                )
                self._change_indent(1)
                check_str += self.indent + "enforce(\n"
                self._change_indent(1)
                frozen_set_element_types_set = set()
                for element_type in element_types:
                    if element_type.startswith("FrozenSet"):
                        frozen_set_element_types_set.add(
                            _get_sub_types_of_compositional_types(element_type)[0]
                        )
                frozen_set_element_types = sorted(frozen_set_element_types_set)
                for frozen_set_element_type in frozen_set_element_types:
                    check_str += (
                        self.indent
                        + "all(type(element) == {} for element in {}) or\n".format(
                            self._to_custom_custom(frozen_set_element_type),
                            content_variable,
                        )
                    )
                check_str = check_str[:-4]
                check_str += "\n"
                self._change_indent(-1)
                if len(frozen_set_element_types) == 1:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for elements of content '{}'. Expected ".format(
                            content_name
                        )
                    )
                    for frozen_set_element_type in frozen_set_element_types:
                        check_str += "'{}'".format(
                            self._to_custom_custom(frozen_set_element_type)
                        )
                    check_str += '.")\n'
                else:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for frozenset elements in content '{}'. Expected either ".format(
                            content_name
                        )
                    )
                    for frozen_set_element_type in frozen_set_element_types:
                        check_str += "'{}' or ".format(
                            self._to_custom_custom(frozen_set_element_type)
                        )
                    check_str = check_str[:-4]
                    check_str += '.")\n'
                self._change_indent(-1)
            if "tuple" in unique_standard_types_list:
                check_str += self.indent + "if type({}) == tuple:\n".format(
                    content_variable
                )
                self._change_indent(1)
                check_str += self.indent + "enforce(\n"
                self._change_indent(1)
                tuple_element_types_set = set()
                for element_type in element_types:
                    if element_type.startswith("Tuple"):
                        tuple_element_types_set.add(
                            _get_sub_types_of_compositional_types(element_type)[0]
                        )
                tuple_element_types = sorted(tuple_element_types_set)
                for tuple_element_type in tuple_element_types:
                    check_str += (
                        self.indent
                        + "all(type(element) == {} for element in {}) or \n".format(
                            self._to_custom_custom(tuple_element_type), content_variable
                        )
                    )
                check_str = check_str[:-4]
                check_str += "\n"
                self._change_indent(-1)
                if len(tuple_element_types) == 1:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for tuple elements in content '{}'. Expected ".format(
                            content_name
                        )
                    )
                    for tuple_element_type in tuple_element_types:
                        check_str += "'{}'".format(
                            self._to_custom_custom(tuple_element_type)
                        )
                    check_str += '.")\n'
                else:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for tuple elements in content '{}'. Expected either ".format(
                            content_name
                        )
                    )
                    for tuple_element_type in tuple_element_types:
                        check_str += "'{}' or ".format(
                            self._to_custom_custom(tuple_element_type)
                        )
                    check_str = check_str[:-4]
                    check_str += '.")\n'
                self._change_indent(-1)
            if "dict" in unique_standard_types_list:
                check_str += self.indent + "if type({}) == dict:\n".format(
                    content_variable
                )
                self._change_indent(1)
                check_str += (
                    self.indent
                    + "for key_of_{}, value_of_{} in {}.items():\n".format(
                        content_name, content_name, content_variable
                    )
                )
                self._change_indent(1)
                check_str += self.indent + "enforce(\n"
                self._change_indent(1)
                dict_key_value_types = dict()
                for element_type in element_types:
                    if element_type.startswith("Dict"):
                        dict_key_value_types[
                            _get_sub_types_of_compositional_types(element_type)[0]
                        ] = _get_sub_types_of_compositional_types(element_type)[1]
                for element1_type in sorted(dict_key_value_types.keys()):
                    check_str += (
                        self.indent
                        + "(type(key_of_{}) == {} and type(value_of_{}) == {}) or\n".format(
                            content_name,
                            self._to_custom_custom(element1_type),
                            content_name,
                            self._to_custom_custom(dict_key_value_types[element1_type]),
                        )
                    )
                check_str = check_str[:-4]
                check_str += "\n"
                self._change_indent(-1)

                if len(dict_key_value_types) == 1:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for dictionary key, value in content '{}'. Expected ".format(
                            content_name
                        )
                    )
                    for key in sorted(dict_key_value_types.keys()):
                        check_str += "'{}', '{}'".format(key, dict_key_value_types[key])
                    check_str += '.")\n'
                else:
                    check_str += (
                        self.indent
                        + ", \"Invalid type for dictionary key, value in content '{}'. Expected ".format(
                            content_name
                        )
                    )
                    for key in sorted(dict_key_value_types.keys()):
                        check_str += "'{}','{}' or ".format(
                            key, dict_key_value_types[key]
                        )
                    check_str = check_str[:-4]
                    check_str += '.")\n'
                self._change_indent(-2)
        elif content_type.startswith("FrozenSet["):
            # check the type
            check_str += (
                self.indent
                + "enforce(type({}) == frozenset, \"Invalid type for content '{}'. Expected 'frozenset'. Found '{{}}'.\".format(type({})))\n".format(
                    content_variable, content_name, content_variable
                )
            )
            element_type = _get_sub_types_of_compositional_types(content_type)[0]
            check_str += self.indent + "enforce(all(\n"
            self._change_indent(1)
            check_str += self.indent + "type(element) == {} for element in {}\n".format(
                self._to_custom_custom(element_type), content_variable
            )
            self._change_indent(-1)
            check_str += (
                self.indent
                + "), \"Invalid type for frozenset elements in content '{}'. Expected '{}'.\")\n".format(
                    content_name, element_type
                )
            )
        elif content_type.startswith("Tuple["):
            # check the type
            check_str += (
                self.indent
                + "enforce(type({}) == tuple, \"Invalid type for content '{}'. Expected 'tuple'. Found '{{}}'.\".format(type({})))\n".format(
                    content_variable, content_name, content_variable
                )
            )
            element_type = _get_sub_types_of_compositional_types(content_type)[0]
            check_str += self.indent + "enforce(all(\n"
            self._change_indent(1)
            check_str += self.indent + "type(element) == {} for element in {}\n".format(
                self._to_custom_custom(element_type), content_variable
            )
            self._change_indent(-1)
            check_str += (
                self.indent
                + "), \"Invalid type for tuple elements in content '{}'. Expected '{}'.\")\n".format(
                    content_name, element_type
                )
            )
        elif content_type.startswith("Dict["):
            # check the type
            check_str += (
                self.indent
                + "enforce(type({}) == dict, \"Invalid type for content '{}'. Expected 'dict'. Found '{{}}'.\".format(type({})))\n".format(
                    content_variable, content_name, content_variable
                )
            )
            element_type_1 = _get_sub_types_of_compositional_types(content_type)[0]
            element_type_2 = _get_sub_types_of_compositional_types(content_type)[1]
            # check the keys type then check the values type
            check_str += (
                self.indent
                + "for key_of_{}, value_of_{} in {}.items():\n".format(
                    content_name, content_name, content_variable
                )
            )
            self._change_indent(1)
            check_str += self.indent + "enforce(\n"
            self._change_indent(1)
            check_str += self.indent + "type(key_of_{}) == {}\n".format(
                content_name, self._to_custom_custom(element_type_1)
            )
            self._change_indent(-1)
            check_str += (
                self.indent
                + ", \"Invalid type for dictionary keys in content '{}'. Expected '{}'. Found '{{}}'.\".format(type(key_of_{})))\n".format(
                    content_name, element_type_1, content_name
                )
            )

            check_str += self.indent + "enforce(\n"
            self._change_indent(1)
            check_str += self.indent + "type(value_of_{}) == {}\n".format(
                content_name, self._to_custom_custom(element_type_2)
            )
            self._change_indent(-1)
            check_str += (
                self.indent
                + ", \"Invalid type for dictionary values in content '{}'. Expected '{}'. Found '{{}}'.\".format(type(value_of_{})))\n".format(
                    content_name, element_type_2, content_name
                )
            )
            self._change_indent(-1)
        else:
            check_str += (
                self.indent
                + "enforce(type({}) == {}, \"Invalid type for content '{}'. Expected '{}'. Found '{{}}'.\".format(type({})))\n".format(
                    content_variable,
                    self._to_custom_custom(content_type),
                    content_name,
                    content_type,
                    content_variable,
                )
            )
        if optional:
            self._change_indent(-1)
        return check_str

    def _message_class_str(self) -> str:
        """
        Produce the content of the Message class.

        :return: the message.py file content
        """
        self._change_indent(0, "s")

        # Header
        cls_str = _copyright_header_str(self.protocol_specification.author) + "\n"

        # Module docstring
        cls_str += (
            self.indent
            + '"""This module contains {}\'s message definition."""\n\n'.format(
                self.protocol_specification.name
            )
        )

        # Imports
        cls_str += self.indent + "import logging\n"
        cls_str += self._import_from_typing_module() + "\n\n"
        cls_str += self.indent + "from aea.configurations.base import ProtocolId\n"
        cls_str += self.indent + "from aea.exceptions import AEAEnforceError, enforce\n"
        cls_str += MESSAGE_IMPORT + "\n"
        if self._import_from_custom_types_module() != "":
            cls_str += "\n" + self._import_from_custom_types_module() + "\n"
        else:
            cls_str += self._import_from_custom_types_module()
        cls_str += (
            self.indent
            + '\n_default_logger = logging.getLogger("aea.packages.{}.protocols.{}.message")\n'.format(
                self.protocol_specification.author, self.protocol_specification.name
            )
        )
        cls_str += self.indent + "\nDEFAULT_BODY_SIZE = 4\n"

        # Class Header
        cls_str += self.indent + "\n\nclass {}Message(Message):\n".format(
            self.protocol_specification_in_camel_case
        )
        self._change_indent(1)
        cls_str += self.indent + '"""{}"""\n\n'.format(
            self.protocol_specification.description
        )

        # Class attributes
        cls_str += (
            self.indent
            + 'protocol_id = ProtocolId.from_str("{}/{}:{}")\n'.format(
                self.protocol_specification.author,
                self.protocol_specification.name,
                self.protocol_specification.version,
            )
        )
        for custom_type in self.spec.all_custom_types:
            cls_str += "\n"
            cls_str += self.indent + "{} = Custom{}\n".format(custom_type, custom_type)

        # Performatives Enum
        cls_str += "\n" + self._performatives_enum_str()

        # __init__
        cls_str += self.indent + "def __init__(\n"
        self._change_indent(1)
        cls_str += self.indent + "self,\n"
        cls_str += self.indent + "performative: Performative,\n"
        cls_str += self.indent + 'dialogue_reference: Tuple[str, str] = ("", ""),\n'
        cls_str += self.indent + "message_id: int = 1,\n"
        cls_str += self.indent + "target: int = 0,\n"
        cls_str += self.indent + "**kwargs,\n"
        self._change_indent(-1)
        cls_str += self.indent + "):\n"
        self._change_indent(1)
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Initialise an instance of {}Message.\n\n".format(
            self.protocol_specification_in_camel_case
        )
        cls_str += self.indent + ":param message_id: the message id.\n"
        cls_str += self.indent + ":param dialogue_reference: the dialogue reference.\n"
        cls_str += self.indent + ":param target: the message target.\n"
        cls_str += self.indent + ":param performative: the message performative.\n"
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "self._performatives = {}\n".format(
            self._performatives_str()
        )
        cls_str += self.indent + "super().__init__(\n"
        self._change_indent(1)
        cls_str += self.indent + "dialogue_reference=dialogue_reference,\n"
        cls_str += self.indent + "message_id=message_id,\n"
        cls_str += self.indent + "target=target,\n"
        cls_str += (
            self.indent
            + "performative={}Message.Performative(performative),\n".format(
                self.protocol_specification_in_camel_case
            )
        )
        cls_str += self.indent + "**kwargs,\n"
        self._change_indent(-1)
        cls_str += self.indent + ")\n"

        self._change_indent(-1)

        # Instance properties
        cls_str += self.indent + "@property\n"
        cls_str += self.indent + "def valid_performatives(self) -> Set[str]:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""Get valid performatives."""\n'
        cls_str += self.indent + "return self._performatives\n\n"
        self._change_indent(-1)
        cls_str += self.indent + "@property\n"
        cls_str += self.indent + "def dialogue_reference(self) -> Tuple[str, str]:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""Get the dialogue_reference of the message."""\n'
        cls_str += (
            self.indent
            + 'enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")\n'
        )
        cls_str += (
            self.indent
            + 'return cast(Tuple[str, str], self.get("dialogue_reference"))\n\n'
        )
        self._change_indent(-1)
        cls_str += self.indent + "@property\n"
        cls_str += self.indent + "def message_id(self) -> int:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""Get the message_id of the message."""\n'
        cls_str += (
            self.indent
            + 'enforce(self.is_set("message_id"), "message_id is not set.")\n'
        )
        cls_str += self.indent + 'return cast(int, self.get("message_id"))\n\n'
        self._change_indent(-1)
        cls_str += self.indent + "@property\n"
        cls_str += (
            self.indent
            + "def performative(self) -> Performative:  # type: ignore # noqa: F821\n"
        )
        self._change_indent(1)
        cls_str += self.indent + '"""Get the performative of the message."""\n'
        cls_str += (
            self.indent
            + 'enforce(self.is_set("performative"), "performative is not set.")\n'
        )
        cls_str += (
            self.indent
            + 'return cast({}Message.Performative, self.get("performative"))\n\n'.format(
                self.protocol_specification_in_camel_case
            )
        )
        self._change_indent(-1)
        cls_str += self.indent + "@property\n"
        cls_str += self.indent + "def target(self) -> int:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""Get the target of the message."""\n'
        cls_str += (
            self.indent + 'enforce(self.is_set("target"), "target is not set.")\n'
        )
        cls_str += self.indent + 'return cast(int, self.get("target"))\n\n'
        self._change_indent(-1)

        for content_name in sorted(self.spec.all_unique_contents.keys()):
            content_type = self.spec.all_unique_contents[content_name]
            cls_str += self.indent + "@property\n"
            cls_str += self.indent + "def {}(self) -> {}:\n".format(
                content_name, self._to_custom_custom(content_type)
            )
            self._change_indent(1)
            cls_str += (
                self.indent
                + '"""Get the \'{}\' content from the message."""\n'.format(
                    content_name
                )
            )
            if not content_type.startswith("Optional"):
                cls_str += (
                    self.indent
                    + 'enforce(self.is_set("{}"), "\'{}\' content is not set.")\n'.format(
                        content_name, content_name
                    )
                )
            cls_str += self.indent + 'return cast({}, self.get("{}"))\n\n'.format(
                self._to_custom_custom(content_type), content_name
            )
            self._change_indent(-1)

        # check_consistency method
        cls_str += self.indent + "def _is_consistent(self) -> bool:\n"
        self._change_indent(1)
        cls_str += (
            self.indent
            + '"""Check that the message follows the {} protocol."""\n'.format(
                self.protocol_specification.name
            )
        )
        cls_str += self.indent + "try:\n"
        self._change_indent(1)
        cls_str += (
            self.indent
            + "enforce(type(self.dialogue_reference) == tuple, \"Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.\""
            ".format(type(self.dialogue_reference)))\n"
        )
        cls_str += (
            self.indent
            + "enforce(type(self.dialogue_reference[0]) == str, \"Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.\""
            ".format(type(self.dialogue_reference[0])))\n"
        )
        cls_str += (
            self.indent
            + "enforce(type(self.dialogue_reference[1]) == str, \"Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.\""
            ".format(type(self.dialogue_reference[1])))\n"
        )
        cls_str += (
            self.indent
            + "enforce(type(self.message_id) == int, \"Invalid type for 'message_id'. Expected 'int'. Found '{}'.\""
            ".format(type(self.message_id)))\n"
        )
        cls_str += (
            self.indent
            + "enforce(type(self.target) == int, \"Invalid type for 'target'. Expected 'int'. Found '{}'.\""
            ".format(type(self.target)))\n\n"
        )

        cls_str += self.indent + "# Light Protocol Rule 2\n"
        cls_str += self.indent + "# Check correct performative\n"
        cls_str += (
            self.indent
            + "enforce(type(self.performative) == {}Message.Performative".format(
                self.protocol_specification_in_camel_case
            )
        )
        cls_str += (
            ", \"Invalid 'performative'. Expected either of '{}'. Found '{}'.\".format("
        )
        cls_str += "self.valid_performatives, self.performative"
        cls_str += "))\n\n"

        cls_str += self.indent + "# Check correct contents\n"
        cls_str += (
            self.indent + "actual_nb_of_contents = len(self.body) - DEFAULT_BODY_SIZE\n"
        )
        cls_str += self.indent + "expected_nb_of_contents = 0\n"
        counter = 1
        for performative, contents in self.spec.speech_acts.items():
            if counter == 1:
                cls_str += self.indent + "if "
            else:
                cls_str += self.indent + "elif "
            cls_str += "self.performative == {}Message.Performative.{}:\n".format(
                self.protocol_specification_in_camel_case, performative.upper(),
            )
            self._change_indent(1)
            nb_of_non_optional_contents = 0
            for content_type in contents.values():
                if not content_type.startswith("Optional"):
                    nb_of_non_optional_contents += 1

            cls_str += self.indent + "expected_nb_of_contents = {}\n".format(
                nb_of_non_optional_contents
            )
            for content_name, content_type in contents.items():
                cls_str += self._check_content_type_str(content_name, content_type)
            counter += 1
            self._change_indent(-1)

        cls_str += "\n"
        cls_str += self.indent + "# Check correct content count\n"
        cls_str += (
            self.indent + "enforce(expected_nb_of_contents == actual_nb_of_contents, "
            '"Incorrect number of contents. Expected {}. Found {}"'
            ".format(expected_nb_of_contents, actual_nb_of_contents))\n\n"
        )

        cls_str += self.indent + "# Light Protocol Rule 3\n"
        cls_str += self.indent + "if self.message_id == 1:\n"
        self._change_indent(1)
        cls_str += (
            self.indent
            + "enforce(self.target == 0, \"Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.\".format(self.target))\n"
        )
        self._change_indent(-1)
        cls_str += self.indent + "else:\n"
        self._change_indent(1)
        cls_str += (
            self.indent + "enforce(0 < self.target < self.message_id, "
            "\"Invalid 'target'. Expected an integer between 1 and {} inclusive. Found {}.\""
            ".format(self.message_id - 1, self.target,))\n"
        )
        self._change_indent(-2)
        cls_str += (
            self.indent + "except (AEAEnforceError, ValueError, KeyError) as e:\n"
        )
        self._change_indent(1)
        cls_str += self.indent + "_default_logger.error(str(e))\n"
        cls_str += self.indent + "return False\n\n"
        self._change_indent(-1)
        cls_str += self.indent + "return True\n"

        return cls_str

    def _valid_replies_str(self) -> str:
        """
        Generate the `valid replies` dictionary.

        :return: the `valid replies` dictionary string
        """
        valid_replies_str = self.indent + "VALID_REPLIES = {\n"
        self._change_indent(1)
        for performative in sorted(self.spec.reply.keys()):
            valid_replies_str += (
                self.indent
                + "{}Message.Performative.{}: frozenset(".format(
                    self.protocol_specification_in_camel_case, performative.upper()
                )
            )
            if len(self.spec.reply[performative]) > 0:
                valid_replies_str += "\n"
                self._change_indent(1)
                valid_replies_str += self.indent + "{"
                for reply in self.spec.reply[performative]:
                    valid_replies_str += "{}Message.Performative.{}, ".format(
                        self.protocol_specification_in_camel_case, reply.upper()
                    )
                valid_replies_str = valid_replies_str[:-2]
                valid_replies_str += "}\n"
                self._change_indent(-1)
            valid_replies_str += self.indent + "),\n"

        self._change_indent(-1)
        valid_replies_str += self.indent + "}"
        return valid_replies_str

    def _end_state_enum_str(self) -> str:
        """
        Generate the end state Enum class.

        :return: the end state Enum string
        """
        enum_str = self.indent + "class EndState(Dialogue.EndState):\n"
        self._change_indent(1)
        enum_str += (
            self.indent
            + '"""This class defines the end states of a {} dialogue."""\n\n'.format(
                self.protocol_specification.name
            )
        )
        tag = 0
        for end_state in self.spec.end_states:
            enum_str += self.indent + "{} = {}\n".format(end_state.upper(), tag)
            tag += 1
        self._change_indent(-1)
        return enum_str

    def _agent_role_enum_str(self) -> str:
        """
        Generate the agent role Enum class.

        :return: the agent role Enum string
        """
        enum_str = self.indent + "class Role(Dialogue.Role):\n"
        self._change_indent(1)
        enum_str += (
            self.indent
            + '"""This class defines the agent\'s role in a {} dialogue."""\n\n'.format(
                self.protocol_specification.name
            )
        )
        for role in self.spec.roles:
            enum_str += self.indent + '{} = "{}"\n'.format(role.upper(), role)
        self._change_indent(-1)
        return enum_str

    def _dialogue_class_str(self) -> str:
        """
        Produce the content of the Message class.

        :return: the message.py file content
        """
        self._change_indent(0, "s")

        # Header
        cls_str = _copyright_header_str(self.protocol_specification.author) + "\n"

        # Module docstring
        cls_str += self.indent + '"""\n'
        cls_str += (
            self.indent
            + "This module contains the classes required for {} dialogue management.\n\n".format(
                self.protocol_specification.name
            )
        )
        cls_str += (
            self.indent
            + "- {}Dialogue: The dialogue class maintains state of a dialogue and manages it.\n".format(
                self.protocol_specification_in_camel_case
            )
        )
        cls_str += (
            self.indent
            + "- {}Dialogues: The dialogues class keeps track of all dialogues.\n".format(
                self.protocol_specification_in_camel_case
            )
        )
        cls_str += self.indent + '"""\n\n'

        # Imports
        cls_str += self.indent + "from abc import ABC\n"
        cls_str += (
            self.indent + "from typing import Callable, FrozenSet, Type, cast\n\n"
        )
        cls_str += self.indent + "from aea.common import Address\n"
        cls_str += self.indent + "from aea.protocols.base import Message\n"
        cls_str += (
            self.indent
            + "from aea.protocols.dialogue.base import Dialogue, DialogueLabel, Dialogues\n\n"
        )
        cls_str += self.indent + "from {}.message import {}Message\n".format(
            self.dotted_path_to_protocol_package,
            self.protocol_specification_in_camel_case,
        )

        # Class Header
        cls_str += "\nclass {}Dialogue(Dialogue):\n".format(
            self.protocol_specification_in_camel_case
        )
        self._change_indent(1)
        cls_str += (
            self.indent
            + '"""The {} dialogue class maintains state of a dialogue and manages it."""\n'.format(
                self.protocol_specification.name
            )
        )

        # Class Constants
        initial_performatives_str = ", ".join(
            [
                "{}Message.Performative.{}".format(
                    self.protocol_specification_in_camel_case, initial_performative
                )
                for initial_performative in self.spec.initial_performatives
            ]
        )
        terminal_performatives_str = ", ".join(
            [
                "{}Message.Performative.{}".format(
                    self.protocol_specification_in_camel_case, terminal_performative
                )
                for terminal_performative in self.spec.terminal_performatives
            ]
        )
        cls_str += (
            self.indent
            + "INITIAL_PERFORMATIVES = frozenset({"
            + initial_performatives_str
            + "})\n"
            + self.indent
            + "TERMINAL_PERFORMATIVES = frozenset({"
            + terminal_performatives_str
            + "})\n"
            + self._valid_replies_str()
        )

        # Enums
        cls_str += "\n" + self._agent_role_enum_str()
        cls_str += "\n" + self._end_state_enum_str()
        cls_str += "\n"

        # initializer
        cls_str += self.indent + "def __init__(\n"
        self._change_indent(1)
        cls_str += self.indent + "self,\n"
        cls_str += self.indent + "dialogue_label: DialogueLabel,\n"
        cls_str += self.indent + "self_address: Address,\n"
        cls_str += self.indent + "role: Dialogue.Role,\n"
        cls_str += self.indent + "message_class: Type[{}Message] = {}Message,\n".format(
            self.protocol_specification_in_camel_case,
            self.protocol_specification_in_camel_case,
        )
        self._change_indent(-1)
        cls_str += self.indent + ") -> None:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Initialize a dialogue.\n\n"
        cls_str += (
            self.indent + ":param dialogue_label: the identifier of the dialogue\n"
        )
        cls_str += (
            self.indent
            + ":param self_address: the address of the entity for whom this dialogue is maintained\n"
        )
        cls_str += (
            self.indent
            + ":param role: the role of the agent this dialogue is maintained for\n"
        )
        cls_str += self.indent + ":return: None\n"
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Dialogue.__init__(\n"
        cls_str += self.indent + "self,\n"
        cls_str += self.indent + "dialogue_label=dialogue_label,\n"
        cls_str += self.indent + "message_class=message_class,\n"
        cls_str += self.indent + "self_address=self_address,\n"
        cls_str += self.indent + "role=role,\n"
        cls_str += self.indent + ")\n"
        self._change_indent(-2)

        # dialogues class
        cls_str += self.indent + "class {}Dialogues(Dialogues, ABC):\n".format(
            self.protocol_specification_in_camel_case
        )
        self._change_indent(1)
        cls_str += (
            self.indent
            + '"""This class keeps track of all {} dialogues."""\n\n'.format(
                self.protocol_specification.name
            )
        )
        end_states_str = ", ".join(
            [
                "{}Dialogue.EndState.{}".format(
                    self.protocol_specification_in_camel_case, end_state.upper()
                )
                for end_state in self.spec.end_states
            ]
        )
        cls_str += self.indent + "END_STATES = frozenset(\n"
        cls_str += self.indent + "{" + end_states_str + "}"
        cls_str += self.indent + ")\n\n"

        cls_str += self.indent + "def __init__(\n"
        self._change_indent(1)
        cls_str += self.indent + "self,\n"
        cls_str += self.indent + "self_address: Address,\n"
        cls_str += (
            self.indent
            + "role_from_first_message: Callable[[Message, Address], Dialogue.Role],\n"
        )
        cls_str += (
            self.indent
            + "dialogue_class: Type[{}Dialogue] = {}Dialogue,\n".format(
                self.protocol_specification_in_camel_case,
                self.protocol_specification_in_camel_case,
            )
        )
        self._change_indent(-1)
        cls_str += self.indent + ") -> None:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Initialize dialogues.\n\n"
        cls_str += (
            self.indent
            + ":param self_address: the address of the entity for whom dialogues are maintained\n"
        )
        cls_str += self.indent + ":return: None\n"
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Dialogues.__init__(\n"
        self._change_indent(1)
        cls_str += self.indent + "self,\n"
        cls_str += self.indent + "self_address=self_address,\n"
        cls_str += (
            self.indent
            + "end_states=cast(FrozenSet[Dialogue.EndState], self.END_STATES),\n"
        )
        cls_str += self.indent + "message_class={}Message,\n".format(
            self.protocol_specification_in_camel_case
        )
        cls_str += self.indent + "dialogue_class=dialogue_class,\n"
        cls_str += self.indent + "role_from_first_message=role_from_first_message,\n"
        self._change_indent(-1)
        cls_str += self.indent + ")\n"
        self._change_indent(-2)
        cls_str += self.indent + "\n"

        return cls_str

    def _custom_types_module_str(self) -> str:
        """
        Produce the contents of the custom_types module, containing classes corresponding to every custom type in the protocol specification.

        :return: the custom_types.py file content
        """
        self._change_indent(0, "s")

        # Header
        cls_str = _copyright_header_str(self.protocol_specification.author) + "\n"

        # Module docstring
        cls_str += '"""This module contains class representations corresponding to every custom type in the protocol specification."""\n'

        # class code per custom type
        for custom_type in self.spec.all_custom_types:
            cls_str += self.indent + "\n\nclass {}:\n".format(custom_type)
            self._change_indent(1)
            cls_str += (
                self.indent
                + '"""This class represents an instance of {}."""\n\n'.format(
                    custom_type
                )
            )
            cls_str += self.indent + "def __init__(self):\n"
            self._change_indent(1)
            cls_str += self.indent + '"""Initialise an instance of {}."""\n'.format(
                custom_type
            )
            cls_str += self.indent + "raise NotImplementedError\n\n"
            self._change_indent(-1)
            cls_str += self.indent + "@staticmethod\n"
            cls_str += (
                self.indent
                + 'def encode({}_protobuf_object, {}_object: "{}") -> None:\n'.format(
                    _camel_case_to_snake_case(custom_type),
                    _camel_case_to_snake_case(custom_type),
                    custom_type,
                )
            )
            self._change_indent(1)
            cls_str += self.indent + '"""\n'
            cls_str += (
                self.indent
                + "Encode an instance of this class into the protocol buffer object.\n\n"
            )
            cls_str += (
                self.indent
                + "The protocol buffer object in the {}_protobuf_object argument is matched with the instance of this class in the '{}_object' argument.\n\n".format(
                    _camel_case_to_snake_case(custom_type),
                    _camel_case_to_snake_case(custom_type),
                )
            )
            cls_str += (
                self.indent
                + ":param {}_protobuf_object: the protocol buffer object whose type corresponds with this class.\n".format(
                    _camel_case_to_snake_case(custom_type)
                )
            )
            cls_str += (
                self.indent
                + ":param {}_object: an instance of this class to be encoded in the protocol buffer object.\n".format(
                    _camel_case_to_snake_case(custom_type)
                )
            )
            cls_str += self.indent + ":return: None\n"
            cls_str += self.indent + '"""\n'
            cls_str += self.indent + "raise NotImplementedError\n\n"
            self._change_indent(-1)

            cls_str += self.indent + "@classmethod\n"
            cls_str += (
                self.indent
                + 'def decode(cls, {}_protobuf_object) -> "{}":\n'.format(
                    _camel_case_to_snake_case(custom_type), custom_type,
                )
            )
            self._change_indent(1)
            cls_str += self.indent + '"""\n'
            cls_str += (
                self.indent
                + "Decode a protocol buffer object that corresponds with this class into an instance of this class.\n\n"
            )
            cls_str += (
                self.indent
                + "A new instance of this class is created that matches the protocol buffer object in the '{}_protobuf_object' argument.\n\n".format(
                    _camel_case_to_snake_case(custom_type)
                )
            )
            cls_str += (
                self.indent
                + ":param {}_protobuf_object: the protocol buffer object whose type corresponds with this class.\n".format(
                    _camel_case_to_snake_case(custom_type)
                )
            )
            cls_str += (
                self.indent
                + ":return: A new instance of this class that matches the protocol buffer object in the '{}_protobuf_object' argument.\n".format(
                    _camel_case_to_snake_case(custom_type)
                )
            )
            cls_str += self.indent + '"""\n'
            cls_str += self.indent + "raise NotImplementedError\n\n"
            self._change_indent(-1)

            cls_str += self.indent + "def __eq__(self, other):\n"
            self._change_indent(1)
            cls_str += self.indent + "raise NotImplementedError\n"
            self._change_indent(-2)
        return cls_str

    def _encoding_message_content_from_python_to_protobuf(
        self, content_name: str, content_type: str,
    ) -> str:
        """
        Produce the encoding of message contents for the serialisation class.

        :param content_name: the name of the content to be encoded
        :param content_type: the type of the content to be encoded

        :return: the encoding string
        """
        encoding_str = ""
        if content_type in PYTHON_TYPE_TO_PROTO_TYPE.keys():
            encoding_str += self.indent + "{} = msg.{}\n".format(
                content_name, content_name
            )
            encoding_str += self.indent + "performative.{} = {}\n".format(
                content_name, content_name
            )
        elif content_type.startswith("FrozenSet") or content_type.startswith("Tuple"):
            encoding_str += self.indent + "{} = msg.{}\n".format(
                content_name, content_name
            )
            encoding_str += self.indent + "performative.{}.extend({})\n".format(
                content_name, content_name
            )
        elif content_type.startswith("Dict"):
            encoding_str += self.indent + "{} = msg.{}\n".format(
                content_name, content_name
            )
            encoding_str += self.indent + "performative.{}.update({})\n".format(
                content_name, content_name
            )
        elif content_type.startswith("Union"):
            sub_types = _get_sub_types_of_compositional_types(content_type)
            for sub_type in sub_types:
                sub_type_name_in_protobuf = _union_sub_type_to_protobuf_variable_name(
                    content_name, sub_type
                )
                encoding_str += self.indent + 'if msg.is_set("{}"):\n'.format(
                    sub_type_name_in_protobuf
                )
                self._change_indent(1)
                encoding_str += self.indent + "performative.{}_is_set = True\n".format(
                    sub_type_name_in_protobuf
                )
                encoding_str += self._encoding_message_content_from_python_to_protobuf(
                    sub_type_name_in_protobuf, sub_type
                )
                self._change_indent(-1)
        elif content_type.startswith("Optional"):
            sub_type = _get_sub_types_of_compositional_types(content_type)[0]
            if not sub_type.startswith("Union"):
                encoding_str += self.indent + 'if msg.is_set("{}"):\n'.format(
                    content_name
                )
                self._change_indent(1)
                encoding_str += self.indent + "performative.{}_is_set = True\n".format(
                    content_name
                )
            encoding_str += self._encoding_message_content_from_python_to_protobuf(
                content_name, sub_type
            )
            if not sub_type.startswith("Union"):
                self._change_indent(-1)
        else:
            encoding_str += self.indent + "{} = msg.{}\n".format(
                content_name, content_name
            )
            encoding_str += self.indent + "{}.encode(performative.{}, {})\n".format(
                content_type, content_name, content_name
            )
        return encoding_str

    def _decoding_message_content_from_protobuf_to_python(
        self,
        performative: str,
        content_name: str,
        content_type: str,
        variable_name_in_protobuf: Optional[str] = "",
    ) -> str:
        """
        Produce the decoding of message contents for the serialisation class.

        :param performative: the performative to which the content belongs
        :param content_name: the name of the content to be decoded
        :param content_type: the type of the content to be decoded
        :param variable_name_in_protobuf: the name of the variable in the protobuf schema

        :return: the decoding string
        """
        decoding_str = ""
        variable_name = (
            content_name
            if variable_name_in_protobuf == ""
            else variable_name_in_protobuf
        )
        if content_type in PYTHON_TYPE_TO_PROTO_TYPE.keys():
            decoding_str += self.indent + "{} = {}_pb.{}.{}\n".format(
                content_name,
                self.protocol_specification.name,
                performative,
                variable_name,
            )
            decoding_str += self.indent + 'performative_content["{}"] = {}\n'.format(
                content_name, content_name
            )
        elif content_type.startswith("FrozenSet"):
            decoding_str += self.indent + "{} = {}_pb.{}.{}\n".format(
                content_name,
                self.protocol_specification.name,
                performative,
                content_name,
            )
            decoding_str += self.indent + "{}_frozenset = frozenset({})\n".format(
                content_name, content_name
            )
            decoding_str += (
                self.indent
                + 'performative_content["{}"] = {}_frozenset\n'.format(
                    content_name, content_name
                )
            )
        elif content_type.startswith("Tuple"):
            decoding_str += self.indent + "{} = {}_pb.{}.{}\n".format(
                content_name,
                self.protocol_specification.name,
                performative,
                content_name,
            )
            decoding_str += self.indent + "{}_tuple = tuple({})\n".format(
                content_name, content_name
            )
            decoding_str += (
                self.indent
                + 'performative_content["{}"] = {}_tuple\n'.format(
                    content_name, content_name
                )
            )
        elif content_type.startswith("Dict"):
            decoding_str += self.indent + "{} = {}_pb.{}.{}\n".format(
                content_name,
                self.protocol_specification.name,
                performative,
                content_name,
            )
            decoding_str += self.indent + "{}_dict = dict({})\n".format(
                content_name, content_name
            )
            decoding_str += (
                self.indent
                + 'performative_content["{}"] = {}_dict\n'.format(
                    content_name, content_name
                )
            )
        elif content_type.startswith("Union"):
            sub_types = _get_sub_types_of_compositional_types(content_type)
            for sub_type in sub_types:
                sub_type_name_in_protobuf = _union_sub_type_to_protobuf_variable_name(
                    content_name, sub_type
                )
                decoding_str += self.indent + "if {}_pb.{}.{}_is_set:\n".format(
                    self.protocol_specification.name,
                    performative,
                    sub_type_name_in_protobuf,
                )
                self._change_indent(1)
                decoding_str += self._decoding_message_content_from_protobuf_to_python(
                    performative=performative,
                    content_name=content_name,
                    content_type=sub_type,
                    variable_name_in_protobuf=sub_type_name_in_protobuf,
                )
                self._change_indent(-1)
        elif content_type.startswith("Optional"):
            sub_type = _get_sub_types_of_compositional_types(content_type)[0]
            if not sub_type.startswith("Union"):
                decoding_str += self.indent + "if {}_pb.{}.{}_is_set:\n".format(
                    self.protocol_specification.name, performative, content_name
                )
                self._change_indent(1)
            decoding_str += self._decoding_message_content_from_protobuf_to_python(
                performative, content_name, sub_type
            )
            if not sub_type.startswith("Union"):
                self._change_indent(-1)
        else:
            decoding_str += self.indent + "pb2_{} = {}_pb.{}.{}\n".format(
                variable_name,
                self.protocol_specification.name,
                performative,
                variable_name,
            )
            decoding_str += self.indent + "{} = {}.decode(pb2_{})\n".format(
                content_name, content_type, variable_name,
            )
            decoding_str += self.indent + 'performative_content["{}"] = {}\n'.format(
                content_name, content_name
            )
        return decoding_str

    def _serialization_class_str(self) -> str:
        """
        Produce the content of the Serialization class.

        :return: the serialization.py file content
        """
        self._change_indent(0, "s")

        # Header
        cls_str = _copyright_header_str(self.protocol_specification.author) + "\n"

        # Module docstring
        cls_str += (
            self.indent
            + '"""Serialization module for {} protocol."""\n\n'.format(
                self.protocol_specification.name
            )
        )

        # Imports
        cls_str += self.indent + "from typing import Any, Dict, cast\n\n"
        cls_str += MESSAGE_IMPORT + "\n"
        cls_str += SERIALIZER_IMPORT + "\n\n"
        cls_str += self.indent + "from {} import (\n    {}_pb2,\n)\n".format(
            self.dotted_path_to_protocol_package, self.protocol_specification.name,
        )
        for custom_type in self.spec.all_custom_types:
            cls_str += (
                self.indent
                + "from {}.custom_types import (\n    {},\n)\n".format(
                    self.dotted_path_to_protocol_package, custom_type,
                )
            )
        cls_str += self.indent + "from {}.message import (\n    {}Message,\n)\n".format(
            self.dotted_path_to_protocol_package,
            self.protocol_specification_in_camel_case,
        )

        # Class Header
        cls_str += self.indent + "\n\nclass {}Serializer(Serializer):\n".format(
            self.protocol_specification_in_camel_case,
        )
        self._change_indent(1)
        cls_str += (
            self.indent
            + '"""Serialization for the \'{}\' protocol."""\n\n'.format(
                self.protocol_specification.name,
            )
        )

        # encoder
        cls_str += self.indent + "@staticmethod\n"
        cls_str += self.indent + "def encode(msg: Message) -> bytes:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Encode a '{}' message into bytes.\n\n".format(
            self.protocol_specification_in_camel_case,
        )
        cls_str += self.indent + ":param msg: the message object.\n"
        cls_str += self.indent + ":return: the bytes.\n"
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "msg = cast({}Message, msg)\n".format(
            self.protocol_specification_in_camel_case
        )
        cls_str += self.indent + "{}_msg = {}_pb2.{}Message()\n".format(
            self.protocol_specification.name,
            self.protocol_specification.name,
            self.protocol_specification_in_camel_case,
        )
        cls_str += self.indent + "{}_msg.message_id = msg.message_id\n".format(
            self.protocol_specification.name
        )
        cls_str += self.indent + "dialogue_reference = msg.dialogue_reference\n"
        cls_str += (
            self.indent
            + "{}_msg.dialogue_starter_reference = dialogue_reference[0]\n".format(
                self.protocol_specification.name
            )
        )
        cls_str += (
            self.indent
            + "{}_msg.dialogue_responder_reference = dialogue_reference[1]\n".format(
                self.protocol_specification.name
            )
        )
        cls_str += self.indent + "{}_msg.target = msg.target\n\n".format(
            self.protocol_specification.name
        )
        cls_str += self.indent + "performative_id = msg.performative\n"
        counter = 1
        for performative, contents in self.spec.speech_acts.items():
            if counter == 1:
                cls_str += self.indent + "if "
            else:
                cls_str += self.indent + "elif "
            cls_str += "performative_id == {}Message.Performative.{}:\n".format(
                self.protocol_specification_in_camel_case, performative.upper()
            )
            self._change_indent(1)
            cls_str += (
                self.indent
                + "performative = {}_pb2.{}Message.{}_Performative()  # type: ignore\n".format(
                    self.protocol_specification.name,
                    self.protocol_specification_in_camel_case,
                    performative.title(),
                )
            )
            for content_name, content_type in contents.items():
                cls_str += self._encoding_message_content_from_python_to_protobuf(
                    content_name, content_type
                )
            cls_str += self.indent + "{}_msg.{}.CopyFrom(performative)\n".format(
                self.protocol_specification.name, performative
            )

            counter += 1
            self._change_indent(-1)
        cls_str += self.indent + "else:\n"
        self._change_indent(1)
        cls_str += (
            self.indent
            + 'raise ValueError("Performative not valid: {}".format(performative_id))\n\n'
        )
        self._change_indent(-1)

        cls_str += self.indent + "{}_bytes = {}_msg.SerializeToString()\n".format(
            self.protocol_specification.name, self.protocol_specification.name
        )
        cls_str += self.indent + "return {}_bytes\n\n".format(
            self.protocol_specification.name
        )
        self._change_indent(-1)

        # decoder
        cls_str += self.indent + "@staticmethod\n"
        cls_str += self.indent + "def decode(obj: bytes) -> Message:\n"
        self._change_indent(1)
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "Decode bytes into a '{}' message.\n\n".format(
            self.protocol_specification_in_camel_case,
        )
        cls_str += self.indent + ":param obj: the bytes object.\n"
        cls_str += self.indent + ":return: the '{}' message.\n".format(
            self.protocol_specification_in_camel_case
        )
        cls_str += self.indent + '"""\n'
        cls_str += self.indent + "{}_pb = {}_pb2.{}Message()\n".format(
            self.protocol_specification.name,
            self.protocol_specification.name,
            self.protocol_specification_in_camel_case,
        )
        cls_str += self.indent + "{}_pb.ParseFromString(obj)\n".format(
            self.protocol_specification.name
        )
        cls_str += self.indent + "message_id = {}_pb.message_id\n".format(
            self.protocol_specification.name
        )
        cls_str += (
            self.indent
            + "dialogue_reference = ({}_pb.dialogue_starter_reference, {}_pb.dialogue_responder_reference)\n".format(
                self.protocol_specification.name, self.protocol_specification.name
            )
        )
        cls_str += self.indent + "target = {}_pb.target\n\n".format(
            self.protocol_specification.name
        )
        cls_str += (
            self.indent
            + 'performative = {}_pb.WhichOneof("performative")\n'.format(
                self.protocol_specification.name
            )
        )
        cls_str += (
            self.indent
            + "performative_id = {}Message.Performative(str(performative))\n".format(
                self.protocol_specification_in_camel_case
            )
        )
        cls_str += (
            self.indent + "performative_content = dict()  # type: Dict[str, Any]\n"
        )
        counter = 1
        for performative, contents in self.spec.speech_acts.items():
            if counter == 1:
                cls_str += self.indent + "if "
            else:
                cls_str += self.indent + "elif "
            cls_str += "performative_id == {}Message.Performative.{}:\n".format(
                self.protocol_specification_in_camel_case, performative.upper()
            )
            self._change_indent(1)
            if len(contents.keys()) == 0:
                cls_str += self.indent + "pass\n"
            else:
                for content_name, content_type in contents.items():
                    cls_str += self._decoding_message_content_from_protobuf_to_python(
                        performative, content_name, content_type
                    )
            counter += 1
            self._change_indent(-1)
        cls_str += self.indent + "else:\n"
        self._change_indent(1)
        cls_str += (
            self.indent
            + 'raise ValueError("Performative not valid: {}.".format(performative_id))\n\n'
        )
        self._change_indent(-1)

        cls_str += self.indent + "return {}Message(\n".format(
            self.protocol_specification_in_camel_case,
        )
        self._change_indent(1)
        cls_str += self.indent + "message_id=message_id,\n"
        cls_str += self.indent + "dialogue_reference=dialogue_reference,\n"
        cls_str += self.indent + "target=target,\n"
        cls_str += self.indent + "performative=performative,\n"
        cls_str += self.indent + "**performative_content\n"
        self._change_indent(-1)
        cls_str += self.indent + ")\n"
        self._change_indent(-2)

        return cls_str

    def _content_to_proto_field_str(
        self, content_name: str, content_type: str, tag_no: int,
    ) -> Tuple[str, int]:
        """
        Convert a message content to its representation in a protocol buffer schema.

        :param content_name: the name of the content
        :param content_type: the type of the content
        :param tag_no: the tag number

        :return: the content in protocol buffer schema and the next tag number to be used
        """
        entry = ""

        if content_type.startswith("FrozenSet") or content_type.startswith(
            "Tuple"
        ):  # it is a <PCT>
            element_type = _get_sub_types_of_compositional_types(content_type)[0]
            proto_type = _python_pt_or_ct_type_to_proto_type(element_type)
            entry = self.indent + "repeated {} {} = {};\n".format(
                proto_type, content_name, tag_no
            )
            tag_no += 1
        elif content_type.startswith("Dict"):  # it is a <PMT>
            key_type = _get_sub_types_of_compositional_types(content_type)[0]
            value_type = _get_sub_types_of_compositional_types(content_type)[1]
            proto_key_type = _python_pt_or_ct_type_to_proto_type(key_type)
            proto_value_type = _python_pt_or_ct_type_to_proto_type(value_type)
            entry = self.indent + "map<{}, {}> {} = {};\n".format(
                proto_key_type, proto_value_type, content_name, tag_no
            )
            tag_no += 1
        elif content_type.startswith("Union"):  # it is an <MT>
            sub_types = _get_sub_types_of_compositional_types(content_type)
            for sub_type in sub_types:
                sub_type_name = _union_sub_type_to_protobuf_variable_name(
                    content_name, sub_type
                )
                content_to_proto_field_str, tag_no = self._content_to_proto_field_str(
                    sub_type_name, sub_type, tag_no
                )
                entry += content_to_proto_field_str
        elif content_type.startswith("Optional"):  # it is an <O>
            sub_type = _get_sub_types_of_compositional_types(content_type)[0]
            content_to_proto_field_str, tag_no = self._content_to_proto_field_str(
                content_name, sub_type, tag_no
            )
            entry = content_to_proto_field_str
            entry += self.indent + "bool {}_is_set = {};\n".format(content_name, tag_no)
            tag_no += 1
        else:  # it is a <CT> or <PT>
            proto_type = _python_pt_or_ct_type_to_proto_type(content_type)
            entry = self.indent + "{} {} = {};\n".format(
                proto_type, content_name, tag_no
            )
            tag_no += 1
        return entry, tag_no

    def _protocol_buffer_schema_str(self) -> str:
        """
        Produce the content of the Protocol Buffers schema.

        :return: the protocol buffers schema content
        """
        self._change_indent(0, "s")

        # heading
        proto_buff_schema_str = self.indent + 'syntax = "proto3";\n\n'
        proto_buff_schema_str += self.indent + "package fetch.aea.{};\n\n".format(
            self.protocol_specification_in_camel_case
        )
        proto_buff_schema_str += self.indent + "message {}Message{{\n\n".format(
            self.protocol_specification_in_camel_case
        )
        self._change_indent(1)

        # custom types
        if (
            (len(self.spec.all_custom_types) != 0)
            and (self.protocol_specification.protobuf_snippets is not None)
            and (self.protocol_specification.protobuf_snippets != "")
        ):
            proto_buff_schema_str += self.indent + "// Custom Types\n"
            for custom_type in self.spec.all_custom_types:
                proto_buff_schema_str += self.indent + "message {}{{\n".format(
                    custom_type
                )
                self._change_indent(1)

                # formatting and adding the custom type protobuf entry
                specification_custom_type = "ct:" + custom_type
                proto_part = self.protocol_specification.protobuf_snippets[
                    specification_custom_type
                ]
                number_of_new_lines = proto_part.count("\n")
                if number_of_new_lines != 0:
                    formatted_proto_part = proto_part.replace(
                        "\n", "\n" + self.indent, number_of_new_lines - 1
                    )
                else:
                    formatted_proto_part = proto_part
                proto_buff_schema_str += self.indent + formatted_proto_part
                self._change_indent(-1)

                proto_buff_schema_str += self.indent + "}\n\n"
            proto_buff_schema_str += "\n"

        # performatives
        proto_buff_schema_str += self.indent + "// Performatives and contents\n"
        for performative, contents in self.spec.speech_acts.items():
            proto_buff_schema_str += self.indent + "message {}_Performative{{".format(
                performative.title()
            )
            self._change_indent(1)
            tag_no = 1
            if len(contents) == 0:
                proto_buff_schema_str += "}\n\n"
                self._change_indent(-1)
            else:
                proto_buff_schema_str += "\n"
                for content_name, content_type in contents.items():
                    (
                        content_to_proto_field_str,
                        tag_no,
                    ) = self._content_to_proto_field_str(
                        content_name, content_type, tag_no
                    )
                    proto_buff_schema_str += content_to_proto_field_str
                self._change_indent(-1)
                proto_buff_schema_str += self.indent + "}\n\n"
        proto_buff_schema_str += "\n"

        # meta-data
        proto_buff_schema_str += self.indent + "// Standard {}Message fields\n".format(
            self.protocol_specification_in_camel_case
        )
        proto_buff_schema_str += self.indent + "int32 message_id = 1;\n"
        proto_buff_schema_str += (
            self.indent + "string dialogue_starter_reference = 2;\n"
        )
        proto_buff_schema_str += (
            self.indent + "string dialogue_responder_reference = 3;\n"
        )
        proto_buff_schema_str += self.indent + "int32 target = 4;\n"
        proto_buff_schema_str += self.indent + "oneof performative{\n"
        self._change_indent(1)
        tag_no = 5
        for performative in self.spec.all_performatives:
            proto_buff_schema_str += self.indent + "{}_Performative {} = {};\n".format(
                performative.title(), performative, tag_no
            )
            tag_no += 1
        self._change_indent(-1)
        proto_buff_schema_str += self.indent + "}\n"
        self._change_indent(-1)

        proto_buff_schema_str += self.indent + "}\n"
        return proto_buff_schema_str

    def _protocol_yaml_str(self) -> str:
        """
        Produce the content of the protocol.yaml file.

        :return: the protocol.yaml content
        """
        protocol_yaml_str = "name: {}\n".format(self.protocol_specification.name)
        protocol_yaml_str += "author: {}\n".format(self.protocol_specification.author)
        protocol_yaml_str += "version: {}\n".format(self.protocol_specification.version)
        protocol_yaml_str += "type: {}\n".format(
            self.protocol_specification.component_type
        )
        protocol_yaml_str += "description: {}\n".format(
            self.protocol_specification.description
        )
        protocol_yaml_str += "license: {}\n".format(self.protocol_specification.license)
        protocol_yaml_str += "aea_version: '{}'\n".format(
            self.protocol_specification.aea_version
        )
        protocol_yaml_str += "fingerprint: {}\n"
        protocol_yaml_str += "fingerprint_ignore_patterns: []\n"
        protocol_yaml_str += "dependencies:\n"
        protocol_yaml_str += "    protobuf: {}\n"

        return protocol_yaml_str

    def _init_str(self) -> str:
        """
        Produce the content of the __init__.py file.

        :return: the __init__.py content
        """
        init_str = _copyright_header_str(self.protocol_specification.author)
        init_str += "\n"
        init_str += '"""This module contains the support resources for the {} protocol."""\n\n'.format(
            self.protocol_specification.name
        )
        init_str += "from {}.message import {}Message\n".format(
            self.dotted_path_to_protocol_package,
            self.protocol_specification_in_camel_case,
        )
        init_str += "from {}.serialization import {}Serializer\n".format(
            self.dotted_path_to_protocol_package,
            self.protocol_specification_in_camel_case,
        )
        init_str += "{}Message.serializer = {}Serializer\n".format(
            self.protocol_specification_in_camel_case,
            self.protocol_specification_in_camel_case,
        )

        return init_str

    def generate_protobuf_only_mode(self) -> None:
        """
        Run the generator in "protobuf only" mode:

        a) validate the protocol specification.
        b) create the protocol buffer schema file.

        :return: None
        """
        # Create the output folder
        output_folder = Path(self.path_to_generated_protocol_package)
        if not output_folder.exists():
            os.mkdir(output_folder)

        # Generate protocol buffer schema file
        _create_protocol_file(
            self.path_to_generated_protocol_package,
            "{}.proto".format(self.protocol_specification.name),
            self._protocol_buffer_schema_str(),
        )

        # Check protobuf schema file is valid
        is_valid_protobuf_schema, msg = check_protobuf_using_protoc(
            self.path_to_generated_protocol_package, self.protocol_specification.name
        )

        if is_valid_protobuf_schema:
            pass
        else:
            # Remove the generated folder and files
            shutil.rmtree(output_folder)
            raise SyntaxError("Error in the protocol buffer schema code:\n" + msg)

    def generate_full_mode(self) -> Optional[str]:
        """
        Run the generator in "full" mode:

        a) validates the protocol specification.
        b) creates the protocol buffer schema file.
        c) generates python modules.
        d) applies black formatting
        e) applies isort formatting

        :return: optional warning message
        """
        # Run protobuf only mode
        self.generate_protobuf_only_mode()

        # Generate Python protocol package
        _create_protocol_file(
            self.path_to_generated_protocol_package, INIT_FILE_NAME, self._init_str()
        )
        _create_protocol_file(
            self.path_to_generated_protocol_package,
            PROTOCOL_YAML_FILE_NAME,
            self._protocol_yaml_str(),
        )
        _create_protocol_file(
            self.path_to_generated_protocol_package,
            MESSAGE_DOT_PY_FILE_NAME,
            self._message_class_str(),
        )
        if (
            self.protocol_specification.dialogue_config is not None
            and self.protocol_specification.dialogue_config != {}
        ):
            _create_protocol_file(
                self.path_to_generated_protocol_package,
                DIALOGUE_DOT_PY_FILE_NAME,
                self._dialogue_class_str(),
            )
        if len(self.spec.all_custom_types) > 0:
            _create_protocol_file(
                self.path_to_generated_protocol_package,
                CUSTOM_TYPES_DOT_PY_FILE_NAME,
                self._custom_types_module_str(),
            )
        _create_protocol_file(
            self.path_to_generated_protocol_package,
            SERIALIZATION_DOT_PY_FILE_NAME,
            self._serialization_class_str(),
        )

        # Run protocol buffer compiler
        try_run_protoc(
            self.path_to_generated_protocol_package, self.protocol_specification.name
        )

        # Run black formatting
        try_run_black_formatting(self.path_to_generated_protocol_package)

        # Run isort formatting
        try_run_isort_formatting(self.path_to_generated_protocol_package)

        # Warn if specification has custom types
        incomplete_generation_warning_msg = None  # type: Optional[str]
        if len(self.spec.all_custom_types) > 0:
            incomplete_generation_warning_msg = "The generated protocol is incomplete, because the protocol specification contains the following custom types: {}. Update the generated '{}' file with the appropriate implementations of these custom types.".format(
                self.spec.all_custom_types, CUSTOM_TYPES_DOT_PY_FILE_NAME
            )
        return incomplete_generation_warning_msg

    def generate(self, protobuf_only: bool = False) -> Optional[str]:
        """
        Run the generator. If in "full" mode (protobuf_only is False), it:

        a) validates the protocol specification.
        b) creates the protocol buffer schema file.
        c) generates python modules.
        d) applies black formatting
        e) applies isort formatting

        If in "protobuf only" mode (protobuf_only is True), it only does a) and b).

        :param protobuf_only: mode of running the generator.
        :return: optional warning message.
        """
        message = None
        if protobuf_only:
            self.generate_protobuf_only_mode()
        else:
            message = self.generate_full_mode()
        return message



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: merkledag.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="merkledag.proto",
    package="merkledag.pb",
    syntax="proto2",
    serialized_options=None,
    serialized_pb=_b(
        '\n\x0fmerkledag.proto\x12\x0cmerkledag.pb"3\n\x06PBLink\x12\x0c\n\x04Hash\x18\x01 \x01(\x0c\x12\x0c\n\x04Name\x18\x02 \x01(\t\x12\r\n\x05Tsize\x18\x03 \x01(\x04";\n\x06PBNode\x12#\n\x05Links\x18\x02 \x03(\x0b\x32\x14.merkledag.pb.PBLink\x12\x0c\n\x04\x44\x61ta\x18\x01 \x01(\x0c'
    ),
)


_PBLINK = _descriptor.Descriptor(
    name="PBLink",
    full_name="merkledag.pb.PBLink",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="Hash",
            full_name="merkledag.pb.PBLink.Hash",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="Name",
            full_name="merkledag.pb.PBLink.Name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="Tsize",
            full_name="merkledag.pb.PBLink.Tsize",
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=33,
    serialized_end=84,
)


_PBNODE = _descriptor.Descriptor(
    name="PBNode",
    full_name="merkledag.pb.PBNode",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="Links",
            full_name="merkledag.pb.PBNode.Links",
            index=0,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="Data",
            full_name="merkledag.pb.PBNode.Data",
            index=1,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=86,
    serialized_end=145,
)

_PBNODE.fields_by_name["Links"].message_type = _PBLINK
DESCRIPTOR.message_types_by_name["PBLink"] = _PBLINK
DESCRIPTOR.message_types_by_name["PBNode"] = _PBNODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PBLink = _reflection.GeneratedProtocolMessageType(
    "PBLink",
    (_message.Message,),
    dict(
        DESCRIPTOR=_PBLINK,
        __module__="merkledag_pb2"
        # @@protoc_insertion_point(class_scope:merkledag.pb.PBLink)
    ),
)
_sym_db.RegisterMessage(PBLink)

PBNode = _reflection.GeneratedProtocolMessageType(
    "PBNode",
    (_message.Message,),
    dict(
        DESCRIPTOR=_PBNODE,
        __module__="merkledag_pb2"
        # @@protoc_insertion_point(class_scope:merkledag.pb.PBNode)
    ),
)
_sym_db.RegisterMessage(PBNode)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test module for Registry publish methods."""

from unittest import TestCase, mock

from click import ClickException

from aea.cli import cli
from aea.cli.publish import (
    _check_is_item_in_local_registry,
    _save_agent_locally,
    _validate_pkp,
)

from tests.conftest import CLI_LOG_OPTION, CliRunner
from tests.test_cli.tools_for_testing import (
    ContextMock,
    PublicIdMock,
    raise_click_exception,
)


@mock.patch("aea.cli.publish.PublicId", PublicIdMock)
@mock.patch("aea.cli.publish._check_is_item_in_local_registry")
@mock.patch("aea.cli.publish.copyfile")
@mock.patch("aea.cli.publish.os.makedirs")
@mock.patch("aea.cli.publish.os.path.exists", return_value=False)
@mock.patch("aea.cli.publish.try_get_item_target_path", return_value="target-dir")
@mock.patch("aea.cli.publish.os.path.join", return_value="joined-path")
@mock.patch("aea.cli.publish.DEFAULT_CONNECTION", "author/default_connection:version")
class SaveAgentLocallyTestCase(TestCase):
    """Test case for _save_agent_locally method."""

    def test_save_agent_locally_positive(
        self,
        path_join_mock,
        try_get_item_target_path_mock,
        path_exists_mock,
        makedirs_mock,
        copyfile_mock,
        _check_is_item_in_local_registry_mock,
    ):
        """Test for save_agent_locally positive result."""
        _save_agent_locally(
            ContextMock(
                connections=["author/default_connection:version", "author/name:version"]
            )
        )
        makedirs_mock.assert_called_once_with("target-dir", exist_ok=True)
        copyfile_mock.assert_called_once_with("joined-path", "joined-path")


class CheckIsItemInLocalRegistryTestCase(TestCase):
    """Test case for _check_is_item_in_local_registry method."""

    @mock.patch("aea.cli.publish.try_get_item_source_path")
    def test__check_is_item_in_local_registry_positive(self, get_path_mock):
        """Test for _check_is_item_in_local_registry positive result."""
        public_id = PublicIdMock.from_str("author/name:version")
        registry_path = "some-registry-path"
        item_type_plural = "items"
        _check_is_item_in_local_registry(public_id, item_type_plural, registry_path)
        get_path_mock.assert_called_once_with(
            registry_path, public_id.author, item_type_plural, public_id.name
        )

    @mock.patch("aea.cli.publish.try_get_item_source_path", raise_click_exception)
    def test__check_is_item_in_local_registry_negative(self):
        """Test for _check_is_item_in_local_registry negative result."""
        public_id = PublicIdMock.from_str("author/name:version")
        registry_path = "some-registry-path"
        item_type_plural = "items"
        with self.assertRaises(ClickException):
            _check_is_item_in_local_registry(public_id, item_type_plural, registry_path)


@mock.patch("aea.cli.utils.decorators._check_aea_project")
@mock.patch("aea.cli.publish._save_agent_locally")
@mock.patch("aea.cli.publish.publish_agent")
@mock.patch("aea.cli.publish._validate_pkp")
@mock.patch("aea.cli.publish._validate_config")
@mock.patch("aea.cli.publish.cast", return_value=ContextMock())
class PublishCommandTestCase(TestCase):
    """Test case for CLI publish command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_publish_positive(self, *mocks):
        """Test for CLI publish positive result."""
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "publish"], standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "publish", "--local"], standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)


class ValidatePkpTestCase(TestCase):
    """Test case for _validate_pkp method."""

    def test__validate_pkp_positive(self):
        """Test _validate_pkp for positive result."""
        private_key_paths = mock.Mock()
        private_key_paths.read_all = mock.Mock(return_value=[])
        _validate_pkp(private_key_paths)
        private_key_paths.read_all.assert_called_once()

    def test__validate_pkp_negative(self):
        """Test _validate_pkp for negative result."""
        private_key_paths = mock.Mock()
        private_key_paths.read_all = mock.Mock(return_value=[1, 2])
        with self.assertRaises(ClickException):
            _validate_pkp(private_key_paths)
        private_key_paths.read_all.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea remove skill` sub-command."""

import os
import shutil
import tempfile
import unittest.mock
from pathlib import Path

import yaml

import aea
import aea.configurations.base
from aea.cli import cli
from aea.configurations.base import AgentConfig, DEFAULT_AEA_CONFIG_FILE

from tests.conftest import AUTHOR, CLI_LOG_OPTION, CliRunner, ROOT_DIR


class TestRemoveSkillWithPublicId:
    """Test that the command 'aea remove skill' works correctly when using the public id."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.skill_id = "fetchai/gym:0.9.0"
        cls.skill_name = "gym"

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)

        # change default registry path
        config = AgentConfig.from_json(yaml.safe_load(open(DEFAULT_AEA_CONFIG_FILE)))
        config.registry_path = os.path.join(ROOT_DIR, "packages")
        yaml.safe_dump(dict(config.json), open(DEFAULT_AEA_CONFIG_FILE, "w"))

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "skill", cls.skill_id],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "remove", "skill", cls.skill_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_zero(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 0

    def test_directory_does_not_exist(self):
        """Test that the directory of the removed skill does not exist."""
        assert not Path("skills", self.skill_name).exists()

    def test_skill_not_present_in_agent_config(self):
        """Test that the name of the removed skill is not present in the agent configuration file."""
        agent_config = aea.configurations.base.AgentConfig.from_json(
            yaml.safe_load(open(DEFAULT_AEA_CONFIG_FILE))
        )
        assert self.skill_id not in agent_config.skills

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestRemoveSkillFailsWhenSkillIsNotSupported:
    """Test that the command 'aea remove skill' fails when the skill is not supported."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.skill_id = "fetchai/gym:0.9.0"

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)

        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "remove", "skill", cls.skill_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_skill_not_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'The skill '{skill_name}' is not supported.'
        """
        s = "The skill '{}' is not supported.".format(self.skill_id)
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestRemoveSkillFailsWhenExceptionOccurs:
    """Test that the command 'aea remove skill' fails when an exception occurs while removing the directory."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.skill_id = "fetchai/gym:0.9.0"
        cls.skill_name = "gym"

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)

        # change default registry path
        config = AgentConfig.from_json(yaml.safe_load(open(DEFAULT_AEA_CONFIG_FILE)))
        config.registry_path = os.path.join(ROOT_DIR, "packages")
        yaml.safe_dump(dict(config.json), open(DEFAULT_AEA_CONFIG_FILE, "w"))

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "skill", cls.skill_id],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        cls.patch = unittest.mock.patch(
            "shutil.rmtree", side_effect=BaseException("an exception")
        )
        cls.patch.start()

        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "remove", "skill", cls.skill_name],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Tests for the webhook connection and channel."""

import asyncio
import json
import logging
from traceback import print_exc
from typing import cast
from unittest.mock import patch

import aiohttp
import pytest
from aiohttp.client_reqrep import ClientResponse

from aea.common import Address
from aea.configurations.base import ConnectionConfig, PublicId
from aea.identity.base import Identity
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue

from packages.fetchai.connections.webhook.connection import WebhookConnection
from packages.fetchai.protocols.http.dialogues import HttpDialogue
from packages.fetchai.protocols.http.dialogues import HttpDialogues as BaseHttpDialogues
from packages.fetchai.protocols.http.message import HttpMessage

from tests.common.mocks import RegexComparator
from tests.conftest import get_host, get_unused_tcp_port


logger = logging.getLogger(__name__)


class HttpDialogues(BaseHttpDialogues):
    """The dialogues class keeps track of all http dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> Dialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return HttpDialogue.Role.CLIENT

        BaseHttpDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


@pytest.mark.asyncio
class TestWebhookConnection:
    """Tests the webhook connection's 'connect' functionality."""

    def setup(self):
        """Initialise the class."""
        self.host = get_host()
        self.port = get_unused_tcp_port()
        self.identity = Identity("", address="some string")
        self.path = "/webhooks/topic/{topic}/"
        self.loop = asyncio.get_event_loop()

        configuration = ConnectionConfig(
            webhook_address=self.host,
            webhook_port=self.port,
            webhook_url_path=self.path,
            connection_id=WebhookConnection.connection_id,
        )
        self.webhook_connection = WebhookConnection(
            configuration=configuration, identity=self.identity,
        )
        self.dialogues = HttpDialogues(self.identity.address)

    async def test_initialization(self):
        """Test the initialisation of the class."""
        assert self.webhook_connection.address == self.identity.address

    @pytest.mark.asyncio
    async def test_connection(self):
        """Test the connect functionality of the webhook connection."""
        await self.webhook_connection.connect()
        assert self.webhook_connection.is_connected is True

    @pytest.mark.asyncio
    async def test_disconnect(self):
        """Test the disconnect functionality of the webhook connection."""
        await self.webhook_connection.connect()
        assert self.webhook_connection.is_connected is True

        await self.webhook_connection.disconnect()
        assert self.webhook_connection.is_connected is False

    def teardown(self):
        """Close connection after testing."""
        try:
            self.loop.run_until_complete(self.webhook_connection.disconnect())
        except Exception:
            print_exc()
            raise

    @pytest.mark.asyncio
    async def test_receive_post_ok(self):
        """Test the connect functionality of the webhook connection."""
        await self.webhook_connection.connect()
        assert self.webhook_connection.is_connected is True
        payload = {"hello": "world"}
        call_task = self.loop.create_task(self.call_webhook("test_topic", json=payload))
        envelope = await asyncio.wait_for(self.webhook_connection.receive(), timeout=10)

        assert envelope

        message = cast(HttpMessage, envelope.message)
        dialogue = self.dialogues.update(message)
        assert dialogue is not None
        assert message.method.upper() == "POST"
        assert message.bodyy.decode("utf-8") == json.dumps(payload)
        await call_task

    @pytest.mark.asyncio
    async def test_send(self):
        """Test the connect functionality of the webhook connection."""
        await self.webhook_connection.connect()
        assert self.webhook_connection.is_connected is True

        http_message = HttpMessage(
            dialogue_reference=("", ""),
            target=0,
            message_id=1,
            performative=HttpMessage.Performative.REQUEST,
            method="get",
            url="/",
            headers="",
            bodyy="",
            version="",
        )
        envelope = Envelope(
            to="addr",
            sender="my_id",
            protocol_id=PublicId.from_str("fetchai/http:0.7.0"),
            message=http_message,
        )
        with patch.object(self.webhook_connection.logger, "warning") as mock_logger:
            await self.webhook_connection.send(envelope)
            await asyncio.sleep(0.01)
            mock_logger.assert_any_call(
                RegexComparator(
                    "Dropping envelope=.* as sending via the webhook is not possible!"
                )
            )

    async def call_webhook(self, topic: str, **kwargs) -> ClientResponse:
        """
        Make a http request to a webhook.

        :param topic: topic to use
        :params **kwargs: data or json for payload

        :return: http response
        """
        path = self.path.format(topic=topic)
        method = kwargs.get("method", "post")
        url = f"http://{self.host}:{self.port}{path}"

        try:
            async with aiohttp.ClientSession() as session:
                async with session.request(method, url, **kwargs) as resp:
                    await resp.read()
                    return resp
        except Exception:
            print_exc()
            raise
